import os
import sys
import logging
from typing import Union, Tuple, List
import pickle
import importlib
import numpy as np
import pandas as pd
import xarray as xr
from abc import ABC, abstractmethod
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import (
    mean_squared_error,
    r2_score,
    mean_absolute_error,
    explained_variance_score,
)
from .logging import get_file_logger
from .operations import (
    normalize,
    denormalize,
    standarize,
    destandarize,
    get_uv_components,
    get_degrees_from_uv,
)


class BlueMathModel(ABC):
    """
    Abstract base class for handling default functionalities across the project.
    """

    @abstractmethod
    def __init__(self) -> None:
        self._logger: logging.Logger = None
        self._exclude_attributes: List[str] = ["_logger"]

    def __getstate__(self):
        """Exclude certain attributes from being pickled."""

        state = self.__dict__.copy()
        for attr in self._exclude_attributes:
            if attr in state:
                del state[attr]
        # Iterate through the state attributes, warning about xr.Datasets
        for key, value in state.items():
            if isinstance(value, xr.Dataset) or isinstance(value, xr.DataArray):
                self.logger.warning(
                    f"Attribute {key} is an xarray Dataset / Dataarray and will be pickled!"
                )

        return state

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = get_file_logger(name=self.__class__.__name__)
        return self._logger

    @logger.setter
    def logger(self, value: logging.Logger) -> None:
        self._logger = value

    def set_logger_name(self, name: str, level: str = "INFO") -> None:
        """Sets the name of the logger."""

        self.logger = get_file_logger(name=name)
        self.logger.setLevel(level)

    def save_model(self, model_path: str, exclude_attributes: List[str] = None) -> None:
        """Saves the model to a file."""

        self.logger.info(f"Saving model to {model_path}")
        if exclude_attributes is not None:
            self._exclude_attributes += exclude_attributes
        with open(model_path, "wb") as f:
            pickle.dump(self, f)

    def load_model(self, model_path: str) -> "BlueMathModel":
        """Loads the model from a file."""

        self.logger.info(f"Loading model from {model_path}")
        with open(model_path, "rb") as f:
            model = pickle.load(f)

        return model

    def list_class_attributes(self) -> list:
        """
        Lists the attributes of the class.

        Returns
        -------
        list
            The attributes of the class.
        """

        return [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def list_class_methods(self) -> list:
        """
        Lists the methods of the class.

        Returns
        -------
        list
            The methods of the class.
        """

        return [
            attr
            for attr in dir(self)
            if callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def check_nans(
        self,
        data: Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset],
        replace_value=None,
        raise_error: bool = False,
    ) -> Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]:
        """
        Checks for NaNs in the data and optionally replaces them.

        Parameters
        ----------
        data : np.ndarray, pd.Series, pd.DataFrame, xr.DataArray or xr.Dataset
            The data to check for NaNs.
        replace_value : any, optional
            The value to replace NaNs with. If None, NaNs will not be replaced.
        raise_error : bool, optional
            Whether to raise an error if NaNs are found. Default is False.

        Returns
        -------
        data : np.ndarray, pd.Series, pd.DataFrame, xr.DataArray or xr.Dataset
            The data with NaNs optionally replaced.

        Raises
        ------
        ValueError
            If NaNs are found and raise_error is True.

        Notes
        -----
        - This method is intended to be used in classes that inherit from the BlueMathModel class.
        - The method checks for NaNs in the data and optionally replaces them with the specified value.

        TODO
        ----
        - Add support for Dask arrays and DataFrames.
        - Add interpolation, moving average, or other methods to replace NaNs.
        """

        if isinstance(data, np.ndarray):
            if np.isnan(data).any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = np.nan_to_num(data, nan=replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
            if data.isnull().values.any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data.fillna(replace_value, inplace=True)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, xr.DataArray) or isinstance(data, xr.Dataset):
            if data.isnull().any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = data.fillna(replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        else:
            self.logger.warning("Data type not supported for NaN check.")

        return data

    def normalize(
        self, data: Union[pd.DataFrame, xr.Dataset], custom_scale_factor: dict = {}
    ) -> Tuple[Union[pd.DataFrame, xr.Dataset], dict]:
        """
        Normalize data to 0-1 using min max scaler approach.
        More info in bluemath_tk.core.operations.normalize.

        Parameters
        ----------
        data : pd.DataFrame or xr.Dataset
            The data to normalize.
        custom_scale_factor : dict, optional
            Custom scale factors for normalization.

        Returns
        -------
        normalized_data : pd.DataFrame or xr.Dataset
            The normalized data.
        scale_factor : dict
            The scale factors used for normalization.
        """

        normalized_data, scale_factor = normalize(
            data=data, custom_scale_factor=custom_scale_factor, logger=self.logger
        )
        return normalized_data, scale_factor

    def denormalize(
        self, normalized_data: pd.DataFrame, scale_factor: dict
    ) -> pd.DataFrame:
        """
        Denormalize data using provided scale_factor.
        More info in bluemath_tk.core.operations.denormalize.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The normalized data to denormalize.
        scale_factor : dict
            The scale factors used for denormalization.

        Returns
        -------
        data : pd.DataFrame
            The denormalized data.
        """

        data = denormalize(normalized_data=normalized_data, scale_factor=scale_factor)
        return data

    def standarize(
        self,
        data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler = None,
    ) -> Tuple[Union[np.ndarray, pd.DataFrame, xr.Dataset], StandardScaler]:
        """
        Standarize data using StandardScaler.
        More info in bluemath_tk.core.operations.standarize.

        Parameters
        ----------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Input data to be standarized.
        scaler : StandardScaler, optional
            Scaler object to use for standarization. Default is None.

        Returns
        -------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data.
        scaler : StandardScaler
            Scaler object used for standarization.
        """

        standarized_data, scaler = standarize(data=data, scaler=scaler)
        return standarized_data, scaler

    def destandarize(
        self,
        standarized_data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler,
    ) -> Union[np.ndarray, pd.DataFrame, xr.Dataset]:
        """
        Destandarize data using provided scaler.
        More info in bluemath_tk.core.operations.destandarize.

        Parameters
        ----------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data to be destandarized.
        scaler : StandardScaler
            Scaler object used for standarization.

        Returns
        -------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Destandarized data.
        """

        data = destandarize(standarized_data=standarized_data, scaler=scaler)
        return data

    @staticmethod
    def get_metrics(
        data1: Union[pd.DataFrame, xr.Dataset],
        data2: Union[pd.DataFrame, xr.Dataset],
    ) -> pd.DataFrame:
        """
        Gets the metrics of the model.

        Parameters
        ----------
        data1 : pd.DataFrame or xr.Dataset
            The first dataset.
        data2 : pd.DataFrame or xr.Dataset
            The second dataset.

        Returns
        -------
        metrics : pd.DataFrame
            The metrics of the model.

        Raises
        ------
        ValueError
            If the DataFrames or Datasets have different shapes.
        TypeError
            If the inputs are not both DataFrames or both xarray Datasets.
        """

        if isinstance(data1, pd.DataFrame) and isinstance(data2, pd.DataFrame):
            if data1.shape != data2.shape:
                raise ValueError("DataFrames must have the same shape")
            variables = data1.columns
        elif isinstance(data1, xr.Dataset) and isinstance(data2, xr.Dataset):
            if sorted(list(data1.dims)) != sorted(list(data2.dims)) or sorted(
                list(data1.data_vars)
            ) != sorted(list(data2.data_vars)):
                raise ValueError(
                    "Datasets must have the same dimensions, coordinates and variables"
                )
            variables = data1.data_vars
        else:
            raise TypeError(
                "Inputs must be either both DataFrames or both xarray Datasets"
            )

        metrics = {}
        for var in variables:
            if isinstance(data1, pd.DataFrame):
                y_true = data1[var]
                y_pred = data2[var]
            else:
                y_true = data1[var].values.reshape(-1)
                y_pred = data2[var].values.reshape(-1)

            metrics[var] = {
                "mean_squared_error": mean_squared_error(y_true, y_pred),
                "r2_score": r2_score(y_true, y_pred),
                "mean_absolute_error": mean_absolute_error(y_true, y_pred),
                "explained_variance_score": explained_variance_score(y_true, y_pred),
            }

        return pd.DataFrame(metrics).T

    @staticmethod
    def get_uv_components(x_deg: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method calculates the u and v components for the given directional data.

        Here, we assume that the directional data is in degrees,
            beign 0° the North direction,
            and increasing clockwise.

                   0° N
                    |
                    |
        270° W <---------> 90° E
                    |
                    |
                  90° S

        Parameters
        ----------
        x_deg : np.ndarray
            The directional data in degrees.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The u and v components.
        """

        return get_uv_components(x_deg)

    @staticmethod
    def get_degrees_from_uv(xu: np.ndarray, xv: np.ndarray) -> np.ndarray:
        """
        This method calculates the degrees from the u and v components.

        Here, we assume u and v represent angles between 0 and 360 degrees,
            where 0° is the North direction,
            and increasing clockwise.

                     (u=0, v=1)
                         |
                         |
        (u=-1, v=0) <---------> (u=1, v=0)
                         |
                         |
                     (u=0, v=-1)

        Parameters
        ----------
        xu : np.ndarray
            The u component.
        xv : np.ndarray
            The v component.

        Returns
        -------
        np.ndarray
            The degrees.
        """

        return get_degrees_from_uv(xu, xv)

    def get_num_processors_available(self) -> int:
        """
        Gets the number of processors available.

        Returns
        -------
        int
            The number of processors available.

        TODO:
        - Check whether available processors are used or not.
        """

        return os.cpu_count()

    def set_num_processors_to_use(self, num_processors: int) -> None:
        """
        Sets the number of processors to use for parallel processing.

        Parameters
        ----------
        num_processors : int
            The number of processors to use.
            If -1, all available processors will be used.

        Raises
        ------
        ValueError
            If the number of processors requested exceeds the number of processors available
        """

        # Retrieve the number of processors available
        num_processors_available = self.get_num_processors_available()

        # Check if the number of processors requested is valid
        if num_processors == -1:
            num_processors = num_processors_available
        elif num_processors <= 0:
            raise ValueError("Number of processors must be greater than 0")
        elif num_processors > num_processors_available:
            raise ValueError(
                f"Number of processors requested ({num_processors}) "
                f"exceeds the number of processors available ({num_processors_available})"
            )

        # Calculate the percentage of processors to use
        percentage = round(num_processors / num_processors_available, 2)
        if percentage < 0.5:
            self.logger.info(
                f"Number of processors requested ({num_processors}) "
                f"is less than 50% of the available processors ({num_processors_available})"
            )
        else:
            self.logger.warning(
                f"Number of processors requested ({num_processors}) "
                f"is more than 50% of the available processors ({num_processors_available})"
            )
        self.logger.info(f"Using {percentage * 100}% of the available processors")
        os.environ["OMP_NUM_THREADS"] = str(num_processors)

        # Re-import numpy if it is already imported
        if "numpy" in sys.modules:
            importlib.reload(np)

    def get_num_processors_used(self) -> int:
        """
        Gets the number of processors used.

        Returns
        -------
        int
            The number of processors used.

        Notes
        -----
        - This method returns the number of processors used by the application.
        - 1 is returned if the number of processors used is not set, as is the case of
          serial processing like Python's built-in functions.
        - Remember that if we run a parallel processing task, the number of processors used
          will be the number of processors set by the task, ehich can be > 1.
          Examples: np.linalg. or numerical models compiled with OpenMP or MPI.
        """

        return int(os.environ.get("OMP_NUM_THREADS", 1))



import os
import copy
import itertools
from typing import List, Union
import subprocess
import numpy as np
import xarray as xr
from jinja2 import Environment, FileSystemLoader
from concurrent.futures import ThreadPoolExecutor, as_completed
from ..core.models import BlueMathModel
from ._utils_wrappers import write_array_in_file, copy_files


class BaseModelWrapper(BlueMathModel):
    """
    Base class for numerical models wrappers.

    Attributes
    ----------
    templates_dir : str
        The directory where the templates are stored.
    model_parameters : dict
        The parameters to be used in the templates.
    output_dir : str
        The directory where the output files will be saved.
    env : Environment
        The Jinja2 environment.
    templates_name : List[str]
        The names of the templates.
    cases_dirs : List[str]
        The list with cases directories.
    cases_context : List[dict]
        The list with cases context.

    Methods
    -------
    _check_parameters_type -> None
        Check if the parameters have the correct type.
    _exec_bash_commands -> None
        Execute bash commands.
    list_available_launchers -> dict
        List the available launchers.
    set_cases_dirs_from_output_dir -> None
        Set the cases directories from the output directory.
    write_array_in_file -> None
        Write an array in a file.
    copy_files -> None
        Copy file(s) from source to destination.
    render_file_from_template -> None
        Render a file from a template.
    create_cases_context_one_by_one -> List[dict]
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.
    create_cases_context_all_combinations -> List[dict]
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.
    build_cases -> None
        Create the cases folders and render the input files.
    run_case -> None
        Run the case based on the launcher specified.
    run_cases -> None
        Run the cases based on the launcher specified.
        Parallel execution is optional.
        Cases to run can be specified.
    run_cases_bulk -> None
        Run the cases based on the launcher specified.
    postprocess_case -> None
        Postprocess the model output for a specific case.
    join_postprocessed_files -> xr.Dataset
        Join the postprocessed files.
    postprocess_cases -> Union[xr.Dataset, List[xr.Dataset]]
        Postprocess the model output.
    """

    def __init__(
        self,
        templates_dir: str,
        model_parameters: dict,
        output_dir: str,
        templates_name: List[str] = "all",
        default_parameters: dict = None,
    ) -> None:
        """
        Initialize the BaseModelWrapper.
        """

        super().__init__()
        if default_parameters is not None:
            self._check_parameters_type(
                default_parameters=default_parameters, model_parameters=model_parameters
            )
        self.templates_dir = templates_dir
        self.model_parameters = model_parameters
        self.output_dir = output_dir
        self._env = Environment(loader=FileSystemLoader(self.templates_dir))
        if templates_name == "all":
            self.logger.warning(
                f"Templates name is 'all', so all templates in {self.templates_dir} will be used."
            )
            self.templates_name = self.env.list_templates()
            self.logger.info(f"Templates names: {self.templates_name}")
        else:
            self.templates_name = templates_name
        self.cases_dirs: List[str] = []
        self.cases_context: List[dict] = []

    @property
    def env(self) -> Environment:
        return self._env

    def _check_parameters_type(
        self, default_parameters: dict, model_parameters: dict
    ) -> None:
        """
        Check if the parameters have the correct type.
        This function is called in the __init__ method of the BaseModelWrapper,
        but default_parameters are defined in the child classes.
        This way, child classes can define default types for parameters.

        Parameters
        ----------
        default_parameters : dict
            The default parameters type for the model.
        model_parameters : dict
            The parameters to be used in the templates.

        Raises
        ------
        ValueError
            If a parameter has the wrong type.
        """

        for model_param, param_value in model_parameters.items():
            if model_param not in default_parameters:
                self.logger.warning(
                    f"Parameter {model_param} is not in the default_parameters"
                )
            else:
                if isinstance(param_value, (list, np.ndarray)) and all(
                    isinstance(item, default_parameters[model_param])
                    for item in param_value
                ):
                    self.logger.info(
                        f"Parameter {model_param} has the correct type: {default_parameters[model_param]}"
                    )
                else:
                    raise ValueError(
                        f"Parameter {model_param} has the wrong type: {default_parameters[model_param]}"
                    )

    @staticmethod
    def _exec_bash_commands(
        str_cmd: str, out_file: str = None, err_file: str = None
    ) -> None:
        """
        Execute bash commands.

        Parameters
        ----------
        str_cmd : str
            The bash command.
        out_file : str, optional
            The name of the output file. If None, the output will be printed in the terminal.
            Default is None.
        err_file : str, optional
            The name of the error file. If None, the error will be printed in the terminal.
            Default is None.
        """

        _stdout = None
        _stderr = None

        if out_file:
            _stdout = open(out_file, "w")
        if err_file:
            _stderr = open(err_file, "w")

        s = subprocess.Popen(str_cmd, shell=True, stdout=_stdout, stderr=_stderr)
        s.wait()

        if out_file:
            _stdout.flush()
            _stdout.close()
        if err_file:
            _stderr.flush()
            _stderr.close()

    def list_available_launchers(self) -> dict:
        """
        List the available launchers.

        Returns
        -------
        dict
            A list with the available launchers.
        """

        if hasattr(self, "available_launchers"):
            return self.available_launchers
        else:
            raise AttributeError("The attribute available_launchers is not defined.")

    def set_cases_dirs_from_output_dir(self) -> None:
        """
        Set the cases directories from the output directory.
        """

        if self.cases_dirs:
            self.logger.warning("Cases directories already set... resetting.")

        self.cases_dirs = sorted(
            [
                os.path.join(self.output_dir, case_dir)
                for case_dir in os.listdir(self.output_dir)
            ]
        )

        self.logger.info(f"Cases directories set from {self.output_dir}.")

    def write_array_in_file(self, array: np.ndarray, filename: str) -> None:
        """
        Write an array in a file.

        Parameters
        ----------
        array : np.ndarray
            The array to be written. Can be 1D or 2D.
        filename : str
            The name of the file.
        """

        write_array_in_file(array=array, filename=filename)

    def copy_files(self, src: str, dst: str) -> None:
        """
        Copy file(s) from source to destination.

        Parameters
        ----------
        src : str
            The source file.
        dst : str
            The destination file.
        """

        copy_files(src=src, dst=dst)

    def render_file_from_template(
        self, template_name: str, context: dict, output_filename: str = None
    ) -> None:
        """
        Render a file from a template.

        Parameters
        ----------
        template_name : str
            The name of the template file.
        context : dict
            The context to be used in the template.
        output_filename : str, optional
            The name of the output file. If None, it will be saved in the output
            directory with the same name as the template.
            Default is None.
        """

        template = self.env.get_template(name=template_name)
        rendered_content = template.render(context)
        if output_filename is None:
            output_filename = os.path.join(self.output_dir, template_name)
        with open(output_filename, "w") as f:
            f.write(rendered_content)

    def create_cases_context_one_by_one(self) -> List[dict]:
        """
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.

        Returns
        -------
        List[dict]
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        num_cases = len(next(iter(self.model_parameters.values())))
        array_of_contexts = []
        for param, values in self.model_parameters.items():
            if len(values) != num_cases:
                raise ValueError(
                    f"All parameters must have the same number of values in one_by_one mode, check {param}"
                )

        for case_num in range(num_cases):
            case_context = {
                param: values[case_num]
                for param, values in self.model_parameters.items()
            }
            array_of_contexts.append(case_context)

        return array_of_contexts

    def create_cases_context_all_combinations(self) -> List[dict]:
        """
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.

        Returns
        -------
        List[dict]
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        keys = self.model_parameters.keys()
        values = self.model_parameters.values()
        combinations = itertools.product(*values)

        array_of_contexts = [
            dict(zip(keys, combination)) for combination in combinations
        ]

        return array_of_contexts

    def build_cases(self, mode: str = "one_by_one") -> None:
        """
        Create the cases folders and render the input files.

        Parameters
        ----------
        mode : str, optional
            The mode to create the cases. Can be "all_combinations" or "one_by_one".
            Default is "one_by_one".
        """

        if mode == "all_combinations":
            self.cases_context = self.create_cases_context_all_combinations()
        elif mode == "one_by_one":
            self.cases_context = self.create_cases_context_one_by_one()
        else:
            raise ValueError(f"Invalid mode to create cases: {mode}")
        for case_num, case_context in enumerate(self.cases_context):
            case_context["case_num"] = case_num
            case_dir = os.path.join(self.output_dir, f"{case_num:04}")
            self.cases_dirs.append(case_dir)
            os.makedirs(case_dir, exist_ok=True)
            for template_name in self.templates_name:
                self.render_file_from_template(
                    template_name=template_name,
                    context=case_context,
                    output_filename=os.path.join(case_dir, template_name),
                )
        self.logger.info(
            f"{len(self.cases_dirs)} cases created in {mode} mode and saved in {self.output_dir}"
        )

    def run_case(
        self,
        case_dir: str,
        launcher: str,
        ouput_log_file: str = "wrapper_out.log",
        error_log_file: str = "wrapper_error.log",
    ) -> None:
        """
        Run the case based on the launcher specified.

        Parameters
        ----------
        case_dir : str
            The case directory.
        launcher : str
            The launcher to run the case.
        ouput_log_file : str, optional
            The name of the output log file. Default is "wrapper_out.log".
        error_log_file : str, optional
            The name of the error log file. Default is "wrapper_error.log".
        """

        # Get launcher command from the available launchers
        launcher = self.list_available_launchers().get(launcher, launcher)

        # Run the case in the case directory
        self.logger.info(f"Running case in {case_dir} with launcher={launcher}.")
        os.chdir(case_dir)
        self._exec_bash_commands(
            str_cmd=launcher,
            out_file=ouput_log_file,
            err_file=error_log_file,
        )

    def run_cases(
        self,
        launcher: str,
        parallel: bool = False,
        cases_to_run: List[int] = None,
    ) -> None:
        """
        Run the cases based on the launcher specified.
        Parallel execution is optional.
        Cases to run can be specified.

        Parameters
        ----------
        launcher : str
            The launcher to run the cases.
        parallel : bool, optional
            If True, the cases will be run in parallel. Default is False.
        cases_to_run : List[int], optional
            The list with the cases to run. Default is None.
        """

        # Get launcher command from the available launchers
        launcher = self.list_available_launchers().get(launcher, launcher)

        if cases_to_run is not None:
            self.logger.warning(
                f"Cases to run was specified, so just {cases_to_run} will be run."
            )
            cases_dir_to_run = [self.cases_dirs[case] for case in cases_to_run]
        else:
            cases_dir_to_run = copy.deepcopy(self.cases_dirs)

        if parallel:
            num_threads = self.get_num_processors_available()
            self.logger.debug(
                f"Running cases in parallel with launcher={launcher}. Number of threads: {num_threads}."
            )
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                future_to_case = {
                    executor.submit(self.run_case, case_dir, launcher): case_dir
                    for case_dir in cases_dir_to_run
                }
                for future in as_completed(future_to_case):
                    case_dir = future_to_case[future]
                    try:
                        future.result()
                    except Exception as exc:
                        self.logger.error(
                            f"Job for {case_dir} generated an exception: {exc}."
                        )
        else:
            self.logger.info(f"Running cases sequentially with launcher={launcher}.")
            for case_dir in cases_dir_to_run:
                try:
                    self.run_case(
                        case_dir=case_dir,
                        launcher=launcher,
                    )
                except Exception as exc:
                    self.logger.error(
                        f"Job for {case_dir} generated an exception: {exc}."
                    )

        if launcher == "docker" or "docker" in launcher:
            # TODO: ALWAYS remove ALL stopped containers after running all cases
            remove_stopped_containers_cmd = 'docker ps -a --filter "ancestor=tausiaj/swash-image:latest" -q | xargs docker rm'
            self._exec_bash_commands(str_cmd=remove_stopped_containers_cmd)

        self.logger.info("All cases ran successfully.")

    def run_cases_bulk(
        self,
        launcher: str,
    ) -> None:
        """
        Run the cases based on the launcher specified.

        Parameters
        ----------
        launcher : str
            The launcher to run the cases.
        """

        self.logger.info(f"Running cases with launcher={launcher}.")
        os.chdir(self.output_dir)
        self._exec_bash_commands(str_cmd=launcher)

    def postprocess_case(self, case_num: int, case_dir: str) -> None:
        """
        Postprocess the model output.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        """

        raise NotImplementedError("The method postprocess_case must be implemented.")

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join the postprocessed files.

        Parameters
        ----------
        postprocessed_files : List[xr.Dataset]
            The list with the postprocessed files.
        """

        raise NotImplementedError(
            "The method join_postprocessed_files must be implemented."
        )

    def postprocess_cases(
        self, cases_to_postprocess: List[int] = None
    ) -> Union[xr.Dataset, List[xr.Dataset]]:
        """
        Postprocess the model output.

        Parameters
        ----------
        cases_to_postprocess : List[int], optional
            The list with the cases to postprocess. Default is None.

        Returns
        -------
        xr.Dataset or List[xr.Dataset]
            The postprocessed file or the list with the postprocessed files.
        """

        # TODO: Check if this option is necessary
        if os.path.exists(os.path.join(self.output_dir, "output_postprocessed.nc")):
            self.logger.warning(
                "Output postprocessed file already exists. Skipping postprocessing."
            )
            return xr.open_dataset(
                os.path.join(self.output_dir, "output_postprocessed.nc")
            )

        if not self.cases_dirs:
            self.logger.warning(
                "Cases directories are not set and will be searched from the output directory."
            )
            self.set_cases_dirs_from_output_dir()

        if cases_to_postprocess is not None:
            self.logger.warning(
                f"Cases to postprocess was specified, so just {cases_to_postprocess} will be postprocessed."
            )
            cases_dir_to_postprocess = [
                self.cases_dirs[case] for case in cases_to_postprocess
            ]
        else:
            cases_to_postprocess = list(range(len(self.cases_dirs)))
            cases_dir_to_postprocess = copy.deepcopy(self.cases_dirs)

        postprocessed_files = []
        for case_num, case_dir in zip(cases_to_postprocess, cases_dir_to_postprocess):
            self.logger.info(f"Postprocessing case {case_num} in {case_dir}.")
            postprocessed_file = self.postprocess_case(
                case_num=case_num, case_dir=case_dir
            )
            postprocessed_files.append(postprocessed_file)

        try:
            return self.join_postprocessed_files(
                postprocessed_files=postprocessed_files
            )
        except NotImplementedError as exc:
            self.logger.error(f"Error joining postprocessed files: {exc}")
            return postprocessed_files



import unittest
import numpy as np
import pandas as pd
from bluemath_tk.datamining.mda import MDA


class TestMDA(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360,
            }
        )
        self.mda = MDA(num_centers=10)

    def test_fit(self):
        self.mda.fit(
            data=self.df,
            directional_variables=["Dir"],
            custom_scale_factor={"Dir": [0, 360]},
            first_centroid_seed=10,
        )
        self.assertIsInstance(self.mda.centroids, pd.DataFrame)
        self.assertEqual(self.mda.centroids.shape[0], 10)

    def test_predict(self):
        data_sample = pd.DataFrame(
            {
                "Hs": np.random.rand(15) * 7,
                "Tp": np.random.rand(15) * 20,
                "Dir": np.random.rand(15) * 360,
            }
        )
        self.mda.fit(
            data=self.df,
            directional_variables=["Dir"],
            custom_scale_factor={},
        )
        nearest_centroids, nearest_centroid_df = self.mda.predict(
            data=data_sample,
        )
        self.assertIsInstance(nearest_centroids, np.ndarray)
        self.assertEqual(len(nearest_centroids), 15)
        self.assertIsInstance(nearest_centroid_df, pd.DataFrame)
        self.assertEqual(nearest_centroid_df.shape[0], 15)

    def test_fit_predict(self):
        nearest_centroids, nearest_centroid_df = self.mda.fit_predict(
            data=self.df,
            directional_variables=["Dir"],
            custom_scale_factor={"Dir": [0, 360]},
        )
        self.assertIsInstance(nearest_centroids, np.ndarray)
        self.assertEqual(len(nearest_centroids), 1000)
        self.assertIsInstance(nearest_centroid_df, pd.DataFrame)
        self.assertEqual(nearest_centroid_df.shape[0], 1000)


if __name__ == "__main__":
    unittest.main()



import xarray as xr
from bluemath_tk.topo_bathy.swan_grid import generate_grid_parameters
from bluemath_tk.wrappers.swan.swan_wrapper import SwanModelWrapper


# Usage example
if __name__ == "__main__":
    # Load GEBCO bathymetry
    bathy_data = xr.open_dataset(
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/bati_tarawa_500m_LONLAT.nc"
    )
    # Generate grid parameters
    grid_parameters = generate_grid_parameters(bathy_data)
    # Define the input parameters
    templates_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/swan/templates/kapi_biwaves/"
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/swan/kapi/"
    # Load swan model parameters
    model_parameters = (
        xr.open_dataset("/home/tausiaj/GitHub-GeoOcean/BlueMath/test_data/subset.nc")
        .to_dataframe()
        .iloc[:100]
        .to_dict(orient="list")
    )
    # Create an instance of the SWAN model wrapper
    swan_wrapper = SwanModelWrapper(
        templates_dir=templates_dir,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    swan_wrapper.build_cases(mode="one_by_one")
    # List available launchers
    print(swan_wrapper.list_available_launchers())
    # Run the model
    swan_wrapper.run_cases(launcher="docker", parallel=True)
    swan_wrapper.run_cases_bulk("sbatch javi_slurm.sh")



from abc import ABC, abstractmethod
from typing import Union
import matplotlib.pyplot as plt
from matplotlib.colors import Colormap
import plotly.graph_objects as go
import cartopy.crs as ccrs


class BasePlotting(ABC):
    """
    Abstract base class for handling default plotting functionalities across the project.
    """

    def __init__(self):
        pass

    @abstractmethod
    def plot_line(self, x, y):
        """
        Abstract method for plotting a line.
        Should be implemented by subclasses.
        """

        pass

    @abstractmethod
    def plot_scatter(self, x, y):
        """
        Abstract method for plotting a scatter plot.
        Should be implemented by subclasses.
        """

        pass

    @abstractmethod
    def plot_map(self, markers=None):
        """
        Abstract method for plotting a map.
        Should be implemented by subclasses.
        """

        pass

    def get_list_of_colors_for_colormap(
        self, cmap: Union[str, Colormap], num_colors: int
    ) -> list:
        """
        Get a list of colors from a colormap.

        Parameters
        ----------
        cmap : str or Colormap
            The colormap to use.
        num_colors : int
            The number of colors to generate.

        Returns
        -------
        list
            A list of colors generated from the colormap.
        """

        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)

        return [cmap(i) for i in range(0, 256, 256 // num_colors)]


class DefaultStaticPlotting(BasePlotting):
    """
    Concrete implementation of BasePlotting with static plotting behaviors.
    """

    # Class-level dictionary for default settings
    templates = {
        "default": {
            "line": {
                "color": "blue",
                "line_style": "-",
            },
            "scatter": {
                "color": "red",
                "size": 10,
                "marker": "o",
            },
        }
    }

    def __init__(self, template: str = "default") -> None:
        """
        Initialize an instance of the DefaultStaticPlotting class.

        Parameters
        ----------
        template : str
            The template to use for the plotting settings. Default is "default".

        Notes
        -----
        - If no keyword arguments are provided, the default template is used.
        - If a keyword argument is provided, it will override the corresponding default setting.
        - Any other provided keyword arguments will be set as instance attributes.
        """

        super().__init__()
        # Update instance attributes with either default template or passed-in values / template
        for key, value in self.templates.get(template, "default").items():
            setattr(self, f"{key}_defaults", value)

    def get_subplots(self, **kwargs):
        fig, ax = plt.subplots(**kwargs)
        return fig, ax

    def plot_line(self, ax, **kwargs):
        c = kwargs.get("c", self.line_defaults.get("color"))
        kwargs.pop("c", None)
        ls = kwargs.get("ls", self.line_defaults.get("line_style"))
        kwargs.pop("ls", None)
        ax.plot(
            c=c,
            ls=ls,
            **kwargs,
        )
        self.set_grid(ax)

    def plot_scatter(self, ax, **kwargs):
        c = kwargs.get("c", self.scatter_defaults.get("color"))
        kwargs.pop("c", None)
        s = kwargs.get("s", self.scatter_defaults.get("size"))
        kwargs.pop("s", None)
        marker = kwargs.get("marker", self.scatter_defaults.get("marker"))
        kwargs.pop("marker", None)
        ax.scatter(
            c=c,
            s=s,
            marker=marker,
            **kwargs,
        )
        self.set_grid(ax)

    def plot_pie(self, ax, **kwargs):
        ax.pie(**kwargs)

    def plot_map(self, ax, **kwargs):
        ax.set_global()
        ax.coastlines()

    def set_title(self, ax, title="Plot Title"):
        """
        Sets the title for a given axis.
        """
        ax.set_title(title)

    def set_xlim(self, ax, xmin, xmax):
        """
        Sets the x-axis limits for a given axis.
        """
        ax.set_xlim(xmin, xmax)

    def set_ylim(self, ax, ymin, ymax):
        """
        Sets the y-axis limits for a given axis.
        """
        ax.set_ylim(ymin, ymax)

    def set_xlabel(self, ax, xlabel="X-axis"):
        """
        Sets the x-axis label for a given axis.
        """
        ax.set_xlabel(xlabel)

    def set_ylabel(self, ax, ylabel="Y-axis"):
        """
        Sets the y-axis label for a given axis.
        """
        ax.set_ylabel(ylabel)

    def set_grid(self, ax, grid=True):
        """
        Sets the grid for a given axis.
        """
        ax.grid(grid)


if __name__ == "__main__":
    static = DefaultStaticPlotting()
    fig, ax = static.get_subplots()
    static.plot_line(ax, x=[1, 2, 3], y=[4, 5, 6])


class DefaultInteractivePlotting(BasePlotting):
    """
    Concrete implementation of BasePlotting with interactive plotting behaviors.
    """

    def __init__(self):
        super().__init__()

    def plot_line(self, x, y):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(x=x, y=y, mode="lines", line=dict(color=self.default_line_color))
        )
        fig.update_layout(
            title="Interactive Line Plot", xaxis_title="X-axis", yaxis_title="Y-axis"
        )
        fig.show()

    def plot_scatter(self, x, y):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=x, y=y, mode="markers", marker=dict(color=self.default_scatter_color)
            )
        )
        fig.update_layout(
            title="Interactive Scatter Plot", xaxis_title="X-axis", yaxis_title="Y-axis"
        )
        fig.show()

    def plot_map(self, markers=None):
        fig = go.Figure(
            go.Scattermapbox(
                lat=[marker[0] for marker in markers] if markers else [],
                lon=[marker[1] for marker in markers] if markers else [],
                mode="markers",
                marker=go.scattermapbox.Marker(size=10, color="red"),
            )
        )
        fig.update_layout(
            mapbox=dict(
                style="open-street-map",
                center=dict(
                    lat=self.default_map_center[0], lon=self.default_map_center[1]
                ),
                zoom=self.default_map_zoom_start,
            ),
            title="Interactive Map with Plotly",
        )
        fig.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import pickle
import time
import sys
import os
import os.path as op
from collections import OrderedDict
from datetime import datetime, date, timedelta

# pip
import numpy as np
import pandas as pd
from sklearn import linear_model
import statsmodels.discrete.discrete_model as sm
import scipy.stats as stat
import xarray as xr

# fix tqdm for notebook
from tqdm import tqdm as tqdm_base


def tqdm(*args, **kwargs):
    if hasattr(tqdm_base, "_instances"):
        for instance in list(tqdm_base._instances):
            tqdm_base._decr_instances(instance)
    return tqdm_base(*args, **kwargs)


# fix library
from scipy import stats

stats.chisqprob = lambda chisq, df: stats.chi2.sf(chisq, df)

# tk
from .io.aux_nc import StoreBugXdset
from .util.time_operations import npdt64todatetime as npdt2dt
from .kma import Persistences
from .plotting.alr import Plot_PValues, Plot_Params, Plot_Terms
from .plotting.wts import (
    Plot_Compare_PerpYear,
    Plot_Compare_Transitions,
    Plot_Compare_Persistences,
)
from .plotting.alr import Plot_Compare_Covariate
from .plotting.alr import Plot_Log_Sim


class ALR_WRP(object):
    "AutoRegressive Logistic Model Wrapper"

    def __init__(self, p_base):
        # data needed for ALR fitting
        self.xds_bmus_fit = None
        self.cluster_size = None

        # ALR terms
        self.d_terms_settings = {}
        self.terms_fit = {}
        self.terms_fit_names = []

        # temporal data storage
        self.mk_order = 0
        self.cov_names = []

        # ALR model core
        self.model = None

        # config (only tested with statsmodels library)
        self.model_library = "statsmodels"  # sklearn / statsmodels

        # paths
        self.p_base = p_base

        # alr model and terms_fit
        self.p_save_model = op.join(p_base, "model.sav")
        self.p_save_terms_fit = op.join(p_base, "terms.sav")

        # store fit and simulated bmus
        self.p_save_fit_xds = op.join(p_base, "xds_input.nc")
        self.p_save_sim_xds = op.join(p_base, "xds_output.nc")

        # export folders for figures
        self.p_report_fit = op.join(p_base, "report_fit")
        self.p_report_sim = op.join(p_base, "report_sim")

        # log sim
        self.p_log_sim_xds = op.join(p_base, "xds_log_sim.nc")

    def SetFitData(self, cluster_size, xds_bmus_fit, d_terms_settings):
        """
        Sets data needed for ALR fitting

        cluster_size - number of clusters in classification
        xds_bmus_fit - xarray.Dataset vars: bmus, dims: time
        d_terms_settings - terms settings. See "SetFittingTerms"
        """

        self.cluster_size = cluster_size
        self.xds_bmus_fit = xds_bmus_fit
        self.SetFittingTerms(d_terms_settings)

        # save bmus series used for fitting
        self.SaveBmus_Fit()

    def SetFittingTerms(self, d_terms_settings):
        "Set terms settings that will be used for fitting"

        # default settings used for ALR terms
        default_settings = {
            "mk_order": 0,
            "constant": False,
            "long_term": False,
            "seasonality": (False, []),
            "covariates": (False, []),
            "covariates_seasonality": (False, []),
        }

        # join user and default input
        for k in default_settings.keys():
            if k not in d_terms_settings:
                d_terms_settings[k] = default_settings[k]

        # generate ALR terms
        bmus_fit = self.xds_bmus_fit.bmus.values
        time_fit = self.xds_bmus_fit.time.values
        cluster_size = self.cluster_size

        self.terms_fit, self.terms_fit_names = self.GenerateALRTerms(
            d_terms_settings, bmus_fit, time_fit, cluster_size, time2yfrac=True
        )

        # store data
        self.mk_order = d_terms_settings["mk_order"]
        self.d_terms_settings = d_terms_settings

    def GenerateALRTerms(
        self, d_terms_settings, bmus, time, cluster_size, time2yfrac=False
    ):
        "Generate ALR terms from user terms settings"

        # terms stored at OrderedDict
        terms = OrderedDict()
        terms_names = []

        # time options (time has to bee yearly fraction)
        if time2yfrac:
            time_yfrac = self.GetFracYears(time)
        else:
            time_yfrac = time

        # constant term
        if d_terms_settings["constant"]:
            terms["constant"] = np.ones((bmus.size, 1))
            terms_names.append("intercept")

        # time term (use custom time array with year decimals)
        if d_terms_settings["long_term"]:
            terms["long_term"] = np.ones((bmus.size, 1))
            terms["long_term"][:, 0] = time_yfrac
            terms_names.append("long_term")

        # seasonality term
        if d_terms_settings["seasonality"][0]:
            phases = d_terms_settings["seasonality"][1]
            temp_seas = np.zeros((len(time_yfrac), 2 * len(phases)))
            c = 0
            for a in phases:
                temp_seas[:, c] = np.cos(a * np.pi * time_yfrac)
                temp_seas[:, c + 1] = np.sin(a * np.pi * time_yfrac)
                terms_names.append("ss_cos_{0}".format(a))
                terms_names.append("ss_sin_{0}".format(a))
                c += 2
            terms["seasonality"] = temp_seas

        # Covariates term
        if d_terms_settings["covariates"][0]:
            # covariates dataset (vars: cov_values, dims: time, cov_names)
            xds_cov = d_terms_settings["covariates"][1]
            cov_names = xds_cov.cov_names.values
            self.cov_names = cov_names  # storage

            # normalize covars
            if not "cov_norm" in xds_cov.keys():
                cov_values = xds_cov.cov_values.values
                cov_norm = (cov_values - cov_values.mean(axis=0)) / cov_values.std(
                    axis=0
                )
            else:
                # simulation covars are previously normalized
                cov_norm = xds_cov.cov_norm.values

            # generate covar terms
            for i in range(cov_norm.shape[1]):
                cn = cov_names[i]
                terms[cn] = np.transpose(np.asmatrix(cov_norm[:, i]))
                terms_names.append(cn)

                # Covariates seasonality
                if d_terms_settings["covariates_seasonality"][0]:
                    cov_season = d_terms_settings["covariates_seasonality"][1]

                    if cov_season[i]:
                        terms["{0}_cos".format(cn)] = np.multiply(
                            terms[cn].T, np.cos(2 * np.pi * time_yfrac)
                        ).T
                        terms["{0}_sin".format(cn)] = np.multiply(
                            terms[cn].T, np.sin(2 * np.pi * time_yfrac)
                        ).T
                        terms_names.append("{0}_cos".format(cn))
                        terms_names.append("{0}_sin".format(cn))

        # markov term
        if d_terms_settings["mk_order"] > 0:
            # dummi for markov chain
            def dummi(csize):
                D = np.ones((csize - 1, csize)) * -1
                for i in range(csize - 1):
                    D[i, csize - 1 - i] = csize - i - 1
                    D[i, csize - 1 + 1 - i :] = 0
                return D

            def dummi_norm(csize):
                D = np.zeros((csize, csize - 1))
                for i in range(csize - 1):
                    D[i, i] = float((csize - i - 1)) / (csize - i)
                    D[i + 1 :, i] = -1.0 / (csize - i)

                return np.transpose(np.flipud(D))

            def helmert_ints(csize, reverse=False):
                D = np.zeros((csize, csize - 1))
                for i in range(csize - 1):
                    D[i, i] = csize - i - 1
                    D[i + 1 :, i] = -1.0

                if reverse:
                    return np.fliplr(np.flipud(D))
                else:
                    return D

            def helmert_norm(csize, reverse=False):
                D = np.zeros((csize, csize - 1))
                for i in range(csize - 1):
                    D[i, i] = float((csize - i - 1)) / (csize - i)
                    D[i + 1 :, i] = -1.0 / (csize - i)

                if reverse:
                    return np.fliplr(np.flipud(D))
                else:
                    return D

            #  helmert
            dum = helmert_norm(cluster_size, reverse=True)

            # solve markov order N
            mk_order = d_terms_settings["mk_order"]
            for i in range(mk_order):
                Z = np.zeros((bmus.size, cluster_size - 1))
                for indz in range(bmus.size - i - 1):
                    Z[indz + i + 1, :] = np.squeeze(dum[bmus[indz] - 1, :])

                terms["markov_{0}".format(i + 1)] = Z

                for ics in range(cluster_size - 1):
                    terms_names.append("mk{0}_{1}".format(i + 1, ics + 1))

        return terms, terms_names

    def GetFracYears(self, time):
        "Returns time in custom year decimal format"

        # fix np.datetime64
        if not "year" in dir(time[0]):
            time_0 = pd.to_datetime(time[0])
            time_1 = pd.to_datetime(time[-1])
            time_d = pd.to_datetime(time[1])
        else:
            time_0 = time[0]
            time_1 = time[-1]
            time_d = time[1]

        # resolution year
        if time_d.year - time_0.year == 1:
            return range(time_1.year - time_0.year + 1)

        # resolution day: get start/end data
        y0 = time_0.year
        m0 = time_0.month
        d0 = time_0.day
        y1 = time_1.year
        m1 = time_1.month
        d1 = time_1.day

        # start "year cicle" at 01/01
        d_y0 = date(y0, 1, 1)

        # time array
        d_0 = date(y0, m0, d0)
        d_1 = date(y1, m1, d1)

        # year_decimal from year start to d1
        delta_y0 = d_1 - d_y0
        y_fraq_y0 = np.array(range(delta_y0.days + 1)) / 365.25

        # cut year_decimal from d_0
        i0 = (d_0 - d_y0).days
        y_fraq = y_fraq_y0[i0:]

        return y_fraq

    def FitModel(self, max_iter=1000):
        "Fits ARL model using sklearn"

        # get fitting data
        X = np.concatenate(list(self.terms_fit.values()), axis=1)
        y = self.xds_bmus_fit.bmus.values

        # fit model
        print("\nFitting autoregressive logistic model ...")
        start_time = time.time()

        if self.model_library == "statsmodels":
            # mount data with pandas
            X = pd.DataFrame(X, columns=self.terms_fit_names)
            y = pd.DataFrame(y, columns=["bmus"])

            # TODO: CAPTURAR LA EVOLUCION DE L (maximun-likelihood)
            self.model = sm.MNLogit(y, X).fit(
                method="lbfgs",
                maxiter=max_iter,
                retall=True,
                full_output=True,
                disp=True,
                warn_convergence=True,
                missing="raise",
            )

        elif self.model_library == "sklearn":
            # use sklearn logistig regression
            self.model = linear_model.LogisticRegression(
                penalty="l2", C=1e5, fit_intercept=False, solver="lbfgs"
            )
            self.model.fit(X, y)

        else:
            print("wrong config: {0} not in model_library".format(self.model_library))
            sys.exit()

        elapsed_time = time.time() - start_time
        print("Optimization done in {0:.2f} seconds\n".format(elapsed_time))

        # save fitted model
        self.SaveModel()

    def SaveModel(self):
        "Saves fitted model (and fitting terms) for future use"

        if not op.isdir(self.p_base):
            os.makedirs(self.p_base)

        # save model
        pickle.dump(self.model, open(self.p_save_model, "wb"))

        # save terms fit
        pickle.dump(
            [self.d_terms_settings, self.terms_fit, self.terms_fit_names],
            open(self.p_save_terms_fit, "wb"),
        )

    def LoadModel(self):
        "Load fitted model (and fitting terms)"

        # load model
        self.model = pickle.load(open(self.p_save_model, "rb"))

        # load terms fit
        self.d_terms_settings, self.terms_fit, self.terms_fit_names = pickle.load(
            open(self.p_save_terms_fit, "rb")
        )

        # load aux data
        if self.d_terms_settings["covariates"][0]:
            cov_names = self.d_terms_settings["covariates"][1].cov_names.values
            self.cov_names = cov_names

        self.mk_order = self.d_terms_settings["mk_order"]

    def SaveBmus_Fit(self):
        "Saves bmus - fit for future use"

        if not op.isdir(self.p_base):
            os.makedirs(self.p_base)

        self.xds_bmus_fit.attrs["cluster_size"] = self.cluster_size
        self.xds_bmus_fit.to_netcdf(self.p_save_fit_xds, "w")

    def LoadBmus_Fit(self):
        "Load bmus - fit"

        self.xds_bmus_fit = xr.open_dataset(self.p_save_fit_xds)
        self.cluster_size = self.xds_bmus_fit.attrs["cluster_size"]

        return self.xds_bmus_fit

    def SaveBmus_Sim(self):
        "Saves bmus - sim for future use"

        if not op.isdir(self.p_base):
            os.makedirs(self.p_base)

        self.xds_bmus_sim.to_netcdf(self.p_save_sim_xds, "w")

    def LoadBmus_Sim(self):
        "Load bmus - sim"

        self.xds_bmus_sim = xr.open_dataset(self.p_save_sim_xds)

        return self.xds_bmus_sim

    def Report_Fit(self, terms_fit=False, summary=False, show=True):
        "Report containing model fitting info"

        # load model
        self.LoadModel()

        # get data
        try:
            pval_df = self.model.pvalues.transpose()
            params_df = self.model.params.transpose()
            name_terms = pval_df.columns.tolist()
        except:
            # TODO: no converge?
            print("warning - statsmodels MNLogit could not provide p-values")
            return

        # output figs
        l_figs = []

        # plot p-values
        f = Plot_PValues(pval_df.values, name_terms, show=show)
        l_figs.append(f)

        # plot parameters
        f = Plot_Params(params_df.values, name_terms, show=show)
        l_figs.append(f)

        # plot terms used for fitting
        if terms_fit:
            f = self.Report_Terms_Fit(p_rep_trms, show=show)
            l_figs.append(f)

        # write summary
        if summary:
            summ = self.model.summary()
            print(summ.as_text())

    def Report_Terms_Fit(self, show=True):
        "Plot terms used for model fitting"

        # load bmus fit
        self.LoadBmus_Fit()

        # get data for plotting
        term_mx = np.concatenate(list(self.terms_fit.values()), axis=1)
        term_ds = [npdt2dt(t) for t in self.xds_bmus_fit.time.values]
        term_ns = self.terms_fit_names

        # Plot terms
        f = Plot_Terms(term_mx, term_ds, term_ns, show=show)
        return f

    def Simulate(
        self,
        num_sims,
        time_sim,
        xds_covars_sim=None,
        log_sim=False,
        overfit_filter=False,
        of_probs=0.98,
        of_pers=5,
    ):
        """
        Launch ARL model simulations

        num_sims           - number of simulations to compute
        time_sim           - time array to solve

        xds_covars_sim     - xr.Dataset (time,), cov_values
            Covariates used at simulation, compatible with "n_sim" dimension
            ("n_sim" dimension (optional) will be iterated with each simulation)

        log_sim            - Store a .nc file with all simulation detailed information.

        filters for exceptional ALR overfit probabilities situation and patch:

        overfit_filter     - overfit filter activation
        of_probs           - overfit filter probabilities activation
        of_pers            - overfit filter persistences activation
        """

        class SimLog(object):
            """
            simulatiom log - records and stores detail info for each time step and n_sim
            """

            def __init__(
                self, time_yfrac, mk_order, num_sims, cluster_size, terms_fit_names
            ):
                # initialize variables to record
                self.terms = np.nan * np.zeros(
                    (
                        len(time_yfrac) - mk_order,
                        num_sims,
                        mk_order + 1,
                        len(terms_fit_names),
                    )
                )
                self.probs = np.nan * np.zeros(
                    (len(time_yfrac) - mk_order, num_sims, mk_order + 1, cluster_size)
                )
                self.ptrns = np.nan * np.zeros(
                    (len(time_yfrac) - mk_order, num_sims, cluster_size)
                )
                self.nrnd = np.nan * np.zeros((len(time_yfrac) - mk_order, num_sims))
                self.evbmu_sims = np.nan * np.zeros(
                    (len(time_yfrac) - mk_order, num_sims)
                )

                # overfit filter variables (filter states and filtered bmus)
                self.of_state = np.zeros(
                    (len(time_yfrac) - mk_order, num_sims), dtype=bool
                )
                self.of_bmus = np.nan * np.zeros((len(time_yfrac) - mk_order, num_sims))

            def Add(self, ix_t, ix_s, terms, prob, probTrans, nrnd, of_state, of_bmus):
                # add iteration to log
                self.terms[ix_t, ix_s, :, :] = terms
                self.probs[ix_t, ix_s, :, :] = prob
                self.ptrns[ix_t, ix_s, :] = probTrans
                self.nrnd[ix_t, ix_s] = nrnd
                self.evbmu_sims[ix_t, ix_s] = np.where(probTrans > nrnd)[0][0] + 1

                # add overfit filter data to log
                self.of_state[ix_t, ix_s] = of_state
                self.of_bmus[ix_t, ix_s] = of_bmus

            def Save(self, p_save, terms_names):
                # use xarray to store netcdf
                xds_log = xr.Dataset(
                    {
                        "alr_terms": (
                            (
                                "time",
                                "n_sim",
                                "mk",
                                "terms",
                            ),
                            self.terms,
                        ),
                        "probs": (("time", "n_sim", "mk", "n_clusters"), self.probs),
                        "probTrans": (("time", "n_sim", "n_clusters"), self.ptrns),
                        "nrnd": (("time", "n_sim"), self.nrnd),
                        "evbmus_sims": (("time", "n_sim"), self.evbmu_sims.astype(int)),
                        "overfit_filter_state": (("time", "n_sim"), self.of_state),
                        "evbmus_sims_filtered": (
                            ("time", "n_sim"),
                            self.of_bmus.astype(int),
                        ),
                    },
                    coords={
                        "time": time_sim[mk_order:],
                        "terms": terms_names,
                    },
                )

                StoreBugXdset(xds_log, p_save)
                print("simulation data log stored at {0}\n".format(p_save))

        class OverfitFilter(object):
            """
            overfit filter for alr outlayer outputs.
            """

            def __init__(self, probs_lim, pers_lim):
                self.active = False
                self.probs_lim = probs_lim
                self.pers_lim = pers_lim
                self.log = ""

            def CheckStatus(self, n_sim, prob, bmus):
                "check current iteration filter status"

                # active filter
                if self.active:
                    # continuation condition
                    self.active = np.nanmax(prob[-1, :]) >= self.probs_lim

                    # log when deactivated
                    if self.active == False:
                        self.log += (
                            "sim. {0:02d} - {1} - deactivated (max prob {2})\n".format(
                                n_sim, time_sim[i], np.nanmax(prob[-1, :])
                            )
                        )

                # inactive filter
                else:
                    # re-activation condition
                    self.active = np.nanmax(prob[-1, :]) >= self.probs_lim and np.all(
                        evbmus[-1 * self.pers_lim :] == new_bmus
                    )

                    # log when activated
                    if self.active:
                        self.log += (
                            "sim. {0:02d} - {1} - activated (max prob {2})\n".format(
                                n_sim, time_sim[i], np.nanmax(prob[-1, :])
                            )
                        )

            def PrintLog(self):
                "Print filter log"

                if self.log != "":
                    print("overfit filter log")
                    print(self.log)

        # switch library probabilities predictor function
        if self.model_library == "statsmodels":
            pred_prob_fun = self.model.predict
        elif self.model_library == "sklearn":
            pred_prob_fun = self.model.predict_proba
        else:
            print("wrong config: {0} not in model_library".format(self.model_library))
            sys.exit()

        # get needed data
        evbmus_values = self.xds_bmus_fit.bmus.values
        time_fit = self.xds_bmus_fit.time.values
        mk_order = self.mk_order

        # times at datetime
        if isinstance(time_sim[0], np.datetime64):
            time_sim = [npdt2dt(t) for t in time_sim]

        # print some info
        tf0 = str(time_fit[0])[:10]
        tf1 = str(time_fit[-1])[:10]
        ts0 = str(time_sim[0])[:10]
        ts1 = str(time_sim[-1])[:10]
        print("ALR model fit   : {0} --- {1}".format(tf0, tf1))
        print("ALR model sim   : {0} --- {1}".format(ts0, ts1))

        # generate time yearly fractional array
        time_yfrac = self.GetFracYears(time_sim)

        # use a d_terms_settigs copy
        d_terms_settings_sim = self.d_terms_settings.copy()

        # initialize optional simulation log
        if log_sim:
            SL = SimLog(
                time_yfrac, mk_order, num_sims, self.cluster_size, self.terms_fit_names
            )

        # initialize ALR overfit filter
        ofilt = OverfitFilter(of_probs, of_pers)

        # initialize ALR simulated bmus array, and overfit filter register array
        evbmus_sims = np.zeros((len(time_yfrac), num_sims))
        ofbmus_sims = np.zeros((len(time_yfrac), num_sims), dtype=bool)

        # start simulations
        print("\nLaunching {0} simulations...\n".format(num_sims))
        for n in range(num_sims):
            # preload some data (simulation covariates)
            cvtxt = ""
            if xds_covars_sim != None:
                # check if n_sim dimension in xds_covars_sim
                if "n_sim" in xds_covars_sim.dims:
                    sim_covars_T = xds_covars_sim.isel(n_sim=n).cov_values.values
                    cvtxt = " (Covs. {0:03d})".format(n + 1)
                else:
                    sim_covars_T = xds_covars_sim.cov_values.values

                sim_covars_T_mean = sim_covars_T.mean(axis=0)
                sim_covars_T_std = sim_covars_T.std(axis=0)

            # progress bar
            pbar = tqdm(
                total=len(time_yfrac) - mk_order,
                file=sys.stdout,
                desc="Sim. Num. {0:03d}{1}".format(n + 1, cvtxt),
            )

            evbmus = evbmus_values[1 : mk_order + 1]
            for i in range(len(time_yfrac) - mk_order):
                # handle simulation covars
                if d_terms_settings_sim["covariates"][0]:
                    # normalize step covars
                    sim_covars_evbmus = sim_covars_T[i : i + mk_order + 1]
                    sim_cov_norm = (
                        sim_covars_evbmus - sim_covars_T_mean
                    ) / sim_covars_T_std

                    # mount step xr.dataset for sim covariates
                    xds_cov_sim_step = xr.Dataset(
                        {
                            "cov_norm": (("time", "cov_names"), sim_cov_norm),
                        },
                        coords={"cov_names": self.cov_names},
                    )

                    d_terms_settings_sim["covariates"] = (True, xds_cov_sim_step)

                # generate time step ALR terms
                terms_i, terms_names = self.GenerateALRTerms(
                    d_terms_settings_sim,
                    np.append(evbmus[i : i + mk_order], 0),
                    time_yfrac[i : i + mk_order + 1],
                    self.cluster_size,
                    time2yfrac=False,
                )

                # Event sequence simulation  (sklearn)
                X = np.concatenate(list(terms_i.values()), axis=1)
                prob = pred_prob_fun(X)  # statsmodels // sklearn functions
                probTrans = np.cumsum(prob[-1, :])

                # generate random cluster with ALR probs
                nrnd = np.random.rand()
                new_bmus = np.where(probTrans > nrnd)[0][0] + 1

                # overfit filter status swich (if active)
                if overfit_filter:
                    ofilt.CheckStatus(n, prob, np.append(evbmus, new_bmus))

                # override overfit bmus if filter active
                if ofilt.active:
                    # criteria: random bmus from that date of the year at  historical
                    ix_of = np.random.choice(
                        np.where(
                            self.xds_bmus_fit["time.dayofyear"]
                            == time_sim[i].timetuple().tm_yday
                        )[0]
                    )
                    new_bmus = self.xds_bmus_fit.bmus.values[ix_of]

                # append_bmus
                evbmus = np.append(evbmus, new_bmus)

                # store overfit filter status
                ofbmus_sims[i + mk_order, n] = ofilt.active

                # optional detail log
                if log_sim:
                    SL.Add(i, n, X, prob, probTrans, nrnd, ofilt.active, new_bmus)

                # update progress bar
                pbar.update(1)

            evbmus_sims[:, n] = evbmus

            # close progress bar
            pbar.close()
        print()  # white line after all progress bars

        # return ALR simulation data in a xr.Dataset
        xds_out = xr.Dataset(
            {
                "evbmus_sims": (("time", "n_sim"), evbmus_sims.astype(int)),
                "ofbmus_sims": (("time", "n_sim"), ofbmus_sims),
            },
            coords={
                "time": time_sim,
            },
        )

        # save output
        StoreBugXdset(xds_out, self.p_save_sim_xds)

        # save log file
        if log_sim:
            SL.Save(self.p_log_sim_xds, terms_names)

        # overfit filter log
        ofilt.PrintLog()

        return xds_out

    def Report_Sim(
        self,
        py_month_ini=1,
        persistences_hists=False,
        persistences_table=False,
        show=True,
    ):
        """
        Report that Compare fitting to simulated bmus

        py_month_ini  - start month for PerpetualYear bmus comparison
        """
        # TODO: add arg n_sim = None (for plotting only one sim output)

        # load fit and sim bmus
        xds_ALR_fit = self.LoadBmus_Fit()
        xds_ALR_sim = self.LoadBmus_Sim()

        # report folder and files
        p_save = self.p_report_sim

        # get data
        cluster_size = self.cluster_size
        bmus_values_sim = xds_ALR_sim.evbmus_sims.values[:]
        bmus_dates_sim = xds_ALR_sim.time.values[:]
        bmus_values_hist = np.reshape(xds_ALR_fit.bmus.values, [-1, 1])
        bmus_dates_hist = xds_ALR_fit.time.values[:]
        num_sims = bmus_values_sim.shape[1]

        # calculate bmus persistences
        pers_hist = Persistences(bmus_values_hist.flatten())
        lsp = [Persistences(bs) for bs in bmus_values_sim.T.astype(int)]
        pers_sim = {k: np.concatenate([x[k] for x in lsp]) for k in lsp[0].keys()}

        # fix datetime 64 dates
        if isinstance(bmus_dates_sim[0], np.datetime64):
            bmus_dates_sim = [npdt2dt(t) for t in bmus_dates_sim]
        if isinstance(bmus_dates_hist[0], np.datetime64):
            bmus_dates_hist = [npdt2dt(t) for t in bmus_dates_hist]

        # output figs
        l_figs = []

        # Plot Perpetual Year (daily) - bmus wt
        fig_PP = Plot_Compare_PerpYear(
            cluster_size,
            bmus_values_sim,
            bmus_dates_sim,
            bmus_values_hist,
            bmus_dates_hist,
            n_sim=num_sims,
            month_ini=py_month_ini,
            show=show,
        )
        l_figs.append(fig_PP)

        # Plot WTs Transition (probability change / scatter Fit vs. ACCUMULATED Sim)
        sttl = "Cluster Probabilities Transitions: All Simulations"
        fig_CT = Plot_Compare_Transitions(
            cluster_size,
            bmus_values_hist,
            bmus_values_sim,
            sttl=sttl,
            show=show,
        )
        l_figs.append(fig_CT)

        # Plot Persistences comparison Fit vs Sim
        if persistences_hists:
            fig_PS = Plot_Compare_Persistences(
                cluster_size,
                pers_hist,
                pers_sim,
                show=show,
            )
            l_figs.append(fig_PS)

        # persistences set table
        if persistences_table:
            print("Persistences by WT (set)")
            for c in range(cluster_size):
                wt = c + 1
                p_h = pers_hist[wt]
                p_s = pers_sim[wt]

                print("WT: {0}".format(wt))
                print("  hist : {0}".format((sorted(set(p_h)))))
                print("  sim. : {0}".format(sorted(set(p_s))))

        # TODO export handling (if show=False)
        # p_save = self.p_report_sim
        # p_rep_PY = None
        # p_rep_VL = None
        # if export:
        #    if not op.isdir(p_save): os.mkdir(p_save)
        #    p_rep_PY = op.join(p_save, 'PerpetualYear.png')
        #    p_rep_VL = op.join(p_save, 'Transitions.png')

        return l_figs

        # TODO: Plot Perpetual Year (monthly)
        # TODO: load covariates if needed for ploting
        # TODO: activate this
        if self.d_terms_settings["covariates"][0]:
            # TODO eliminar tiempos duplicados
            time_hist_covars = bmus_dates_hist
            time_sim_covars = bmus_dates_sim

            # covars fit
            xds_cov_fit = self.d_terms_settings["covariates"][1]
            cov_names = xds_cov_fit.cov_names.values
            cov_fit_values = xds_cov_fit.cov_values.values

            # covars sim
            cov_sim_values = xds_cov_sim.cov_values.values

            for ic, cn in enumerate(cov_names):
                # get covariate data
                cf_val = cov_fit_values[:, ic]
                cs_val = cov_sim_values[:, ic]

                # plot covariate - bmus wt
                p_rep_cn = op.join(p_save, "{0}_comp.png".format(cn))
                Plot_Compare_Covariate(
                    cluster_size,
                    bmus_values_sim,
                    bmus_dates_sim,
                    bmus_values_hist,
                    bmus_dates_hist,
                    cs_val,
                    time_sim_covars,
                    cf_val,
                    time_hist_covars,
                    cn,
                    n_sim=num_sims,
                    p_export=p_rep_cn,
                )

    def Report_Sim_Log(self, n_sim=0, t_slice=None, show=True):
        """
        Interactive plot for simulation log

        n_sim  - simulation log to plot
        """

        # load fit and sim bmus
        xds_log = xr.open_dataset(self.p_log_sim_xds, decode_times=True)

        # get simulation
        log_sim = xds_log.isel(n_sim=n_sim)

        if t_slice != None:
            log_sim = log_sim.sel(time=t_slice)

        # plot interactive report
        Plot_Log_Sim(log_sim)



import numpy as np
import xarray as xr


def get_2d_dataset():
    # Define the coordinates
    coord1 = np.linspace(-100, 100, 20)
    coord2 = np.linspace(-100, 100, 20)
    coord3 = np.arange(1, 50)

    # Create a meshgrid
    coord1, coord2, coord3 = np.meshgrid(coord1, coord2, coord3, indexing="ij")

    # Create a 3D dataset
    X = (
        np.sin(np.radians(coord1)) * np.cos(np.radians(coord2)) * np.sin(coord3)
        + np.sin(2 * np.radians(coord1))
        * np.cos(2 * np.radians(coord2))
        * np.sin(2 * coord3)
        + np.sin(3 * np.radians(coord1))
        * np.cos(3 * np.radians(coord2))
        * np.sin(3 * coord3)
    )
    # Create a 3D dataset
    Y = -np.sin(X)

    # Create an xarray dataset
    ds = xr.Dataset(
        {
            "X": (["coord1", "coord2", "coord3"], X),
            "Y": (["coord1", "coord2", "coord3"], Y),
        },
        coords={
            "coord1": coord1[:, 0, 0],
            "coord2": coord2[0, :, 0],
            "coord3": coord3[0, 0, :],
        },
    )

    return ds



import unittest
from bluemath_tk.core.data.sample_data import get_2d_dataset
from bluemath_tk.datamining.pca import PCA


class TestPCA(unittest.TestCase):
    def setUp(self):
        self.ds = get_2d_dataset()
        self.pca = PCA(n_components=5, debug=True)
        self.ipca = PCA(n_components=5, is_incremental=True)

    def test_fit_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
            windows_in_pca_dim_for_rows={"X": [3], "Y": [1]},
            value_to_replace_nans={"X": 0.0, "X_3": 1.0, "Y": 0.0},
            nan_threshold_to_drop={"X": 0.5, "Y": 0.5},
        )
        self.assertEqual(self.pca.is_fitted, True)
        self.assertEqual(pcs.PCs.shape[1], 5)
        self.assertEqual(pcs.PCs.shape[0], self.ds.sizes["coord3"])
        self.assertCountEqual(self.pca.eofs.data_vars, ["X", "X_3", "Y", "Y_1"])

    def test_inverse_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        reconstructed_ds = self.pca.inverse_transform(PCs=pcs)
        self.assertAlmostEqual(
            self.ds.isel(coord1=5, coord2=5, coord3=5),
            reconstructed_ds.isel(coord1=5, coord2=5, coord3=5),
        )

    def test_incremental_fit(self):
        self.ipca.fit(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        self.assertEqual(self.ipca.is_fitted, True)


if __name__ == "__main__":
    unittest.main()



import time
from typing import List, Tuple, Callable
import numpy as np
import pandas as pd
from scipy.optimize import fminbound, fmin
from concurrent.futures import ThreadPoolExecutor, as_completed
from ._base_interpolation import BaseInterpolation
from ..core.decorators import validate_data_rbf


def gaussian_kernel(r: float, const: float) -> float:
    """
    This function calculates the Gaussian kernel for the given distance and constant.

    Parameters
    ----------
    r : float
        The distance.
    const : float
        The constant (default name is usually sigma for gaussian kernel).

    Returns
    -------
    float
        The Gaussian kernel value.

    Notes
    -----
    - The Gaussian kernel is defined as:
      K(r) = exp(r^2 / 2 * const^2) (https://en.wikipedia.org/wiki/Gaussian_function)
    - Here, we are assuming the mean is 0.
    """

    return np.exp(-0.5 * r * r / (const * const))


def multiquadratic_kernel(r, const):
    return np.sqrt(1 + (r / const) ** 2)


def inverse_kernel(r, const):
    return 1 / np.sqrt(1 + (r / const) ** 2)


def cubic_kernel(r, const):
    return r**3


def thin_plate_kernel(r, const):
    return r**2 * np.log(r / const)


class RBFError(Exception):
    """
    Custom exception for RBF class.
    """

    def __init__(self, message: str = "RBF error occurred."):
        self.message = message
        super().__init__(self.message)


class RBF(BaseInterpolation):
    """
    Radial Basis Function (RBF) interpolation model.

    Attributes
    ----------
    sigma_min : float
        The minimum value for the sigma parameter.
        This value might change in the optimization process.
    sigma_max : float
        The maximum value for the sigma parameter.
        This value might change in the optimization process.
    sigma_diff : float
        The difference between the optimal bounded sigma and the minimum and maximum sigma values.
        If the difference is less than this value, the optimization process continues.
    kernel : str
        The kernel to use for the RBF model.
        The available kernels are:

            - gaussian              : ``exp(-1/2 * (r / const)**2)``
            - multiquadratic        : ``sqrt(1 + (r / const)**2)``
            - inverse               : ``1 / sqrt(1 + (r / const)**2)``
            - cubic                 : ``r**3``
            - thin_plate            : ``r**2 * log(r / const)``

    kernel_func : function
        The kernel function to use for the RBF model.
    smooth : float
        The smoothness parameter.
    subset_data : pd.DataFrame
        The subset data used to fit the model.
    normalized_subset_data : pd.DataFrame
        The normalized subset data used to fit the model.
    target_data : pd.DataFrame
        The target data used to fit the model.
    normalized_target_data : pd.DataFrame
        The normalized target data used to fit the model.
        This attribute is only set if normalize_target_data is True in the fit method.
    subset_directional_variables : List[str]
        The subset directional variables.
    target_directional_variables : List[str]
        The target directional variables.
    subset_processed_variables : List[str]
        The subset processed variables.
    target_processed_variables : List[str]
        The target processed variables.
    subset_custom_scale_factor : dict
        The custom scale factor for the subset data.
    target_custom_scale_factor : dict
        The custom scale factor for the target data.
    subset_scale_factor : dict
        The scale factor for the subset data.
    target_scale_factor : dict
        The scale factor for the target data.
    rbf_coeffs : pd.DataFrame
        The RBF coefficients for the target variables.
    opt_sigmas : dict
        The optimal sigmas for the target variables.

    Methods
    -------
    fit :
        Fits the model to the data.
    predict :
        Predicts the data for the provided dataset.
    fit_predict :
        Fits the model to the subset and predicts the interpolated dataset.

    Notes
    -----
    TODO: For the moment, this class only supports optimization for one parameter kernels.
          For this reason, we only have sigma as the parameter to optimize.
          This sigma refers to the sigma parameter in the Gaussian kernel (but is used for all kernels).

    Main reference for sigma optimization: https://link.springer.com/article/10.1023/A:1018975909870

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from bluemath_tk.interpolation.rbf import RBF
    >>> dataset = pd.DataFrame(
    ...     {
    ...         "Hs": np.random.rand(1000) * 7,
    ...         "Tp": np.random.rand(1000) * 20,
    ...         "Dir": np.random.rand(1000) * 360,
    ...     }
    ... )
    >>> subset = dataset.sample(frac=0.25)
    >>> target = pd.DataFrame(
    ...     {
    ...         "HsPred": self.subset["Hs"] * 2 + self.subset["Tp"] * 3,
    ...         "DirPred": - self.subset["Dir"],
    ...     }
    ... )
    >>> rbf = RBF()
    >>> predictions = rbf.fit_predict(
    ...     subset_data=subset,
    ...     subset_directional_variables=["Dir"],
    ...     target_data=target,
    ...     target_directional_variables=["DirPred"],
    ...     normalize_target_data=True,
    ...     dataset=dataset,
    ...     num_threads=4,
    ...     iteratively_update_sigma=True,
    ... )
    """

    rbf_kernels = {
        "gaussian": gaussian_kernel,
        "multiquadratic": multiquadratic_kernel,
        "inverse": inverse_kernel,
        "cubic": cubic_kernel,
        "thin_plate": thin_plate_kernel,
    }

    def __init__(
        self,
        sigma_min: float = 0.001,
        sigma_max: float = 0.1,
        sigma_diff: float = 0.0001,
        sigma_opt: float = None,
        kernel: str = "gaussian",
        smooth: float = 1e-5,
    ):
        """
        Raises
        ------
        ValueError
            If the sigma_min is not a positive float.
            If the sigma_max is not a positive float greater than sigma_min.
            If the sigma_diff is not a positive float.
            If the kernel is not a string and one of the available kernels.
            If the smooth is not a positive float.
        """

        initial_msg = f"""
        ---------------------------------------------------------------------------------
        | Initializing RBF interpolation model with the following parameters:
        |    - sigma_min: {sigma_min}
        |    - sigma_max: {sigma_max}
        |    - sigma_diff: {sigma_diff}
        |    - sigma_opt: {sigma_opt}
        |    - kernel: {kernel}
        |    - smooth: {smooth}
        | For more information, please refer to the documentation.
        | Recommended lecture: https://link.springer.com/article/10.1023/A:1018975909870
        ---------------------------------------------------------------------------------
        """
        print(initial_msg)

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if not isinstance(sigma_min, float) or sigma_min < 0:
            raise ValueError("sigma_min must be a positive float.")
        self._sigma_min = sigma_min
        if not isinstance(sigma_max, float) or sigma_max < sigma_min:
            raise ValueError(
                "sigma_max must be a positive float greater than sigma_min."
            )
        self._sigma_max = sigma_max
        if not isinstance(sigma_diff, float) or sigma_diff < 0:
            raise ValueError("sigma_diff must be a positive float.")
        self._sigma_diff = sigma_diff
        if not isinstance(kernel, str) or kernel not in self.rbf_kernels.keys():
            raise ValueError(
                f"kernel must be a string and one of {list(self.rbf_kernels.keys())}."
            )
        if sigma_opt is not None:
            if not isinstance(sigma_opt, float) or sigma_opt < 0:
                raise ValueError("sigma_opt must be a positive float.")
        self._sigma_opt = sigma_opt
        if not isinstance(kernel, str) or kernel not in self.rbf_kernels.keys():
            raise ValueError(
                f"kernel must be a string and one of {list(self.rbf_kernels.keys())}."
            )
        self._kernel = kernel
        self._kernel_func = self.rbf_kernels[self.kernel]
        if not isinstance(smooth, float) or smooth < 0:
            raise ValueError("smooth must be a positive float.")
        self._smooth = smooth
        # Below, we initialize the attributes that will be set in the fit method
        self.is_fitted: bool = False
        self.is_target_normalized: bool = False
        self._subset_data: pd.DataFrame = pd.DataFrame()
        self._normalized_subset_data: pd.DataFrame = pd.DataFrame()
        self._target_data: pd.DataFrame = pd.DataFrame()
        self._normalized_target_data: pd.DataFrame = pd.DataFrame()
        self._subset_directional_variables: List[str] = []
        self._target_directional_variables: List[str] = []
        self._subset_processed_variables: List[str] = []
        self._target_processed_variables: List[str] = []
        self._subset_custom_scale_factor: dict = {}
        self._target_custom_scale_factor: dict = {}
        self._subset_scale_factor: dict = {}
        self._target_scale_factor: dict = {}
        self._rbf_coeffs: pd.DataFrame = pd.DataFrame()
        self._opt_sigmas: dict = {}

    @property
    def sigma_min(self) -> float:
        return self._sigma_min

    @property
    def sigma_max(self) -> float:
        return self._sigma_max

    @property
    def sigma_diff(self) -> float:
        return self._sigma_diff

    @property
    def sigma_opt(self) -> float:
        return self._sigma_opt

    @property
    def kernel(self) -> str:
        return self._kernel

    @property
    def kernel_func(self) -> Callable:
        return self._kernel_func

    @property
    def smooth(self) -> float:
        return self._smooth

    @property
    def subset_data(self) -> pd.DataFrame:
        return self._subset_data

    @property
    def normalized_subset_data(self) -> pd.DataFrame:
        return self._normalized_subset_data

    @property
    def target_data(self) -> pd.DataFrame:
        return self._target_data

    @property
    def normalized_target_data(self) -> pd.DataFrame:
        if self._normalized_target_data.empty:
            raise ValueError("Target data is not normalized.")
        return self._normalized_target_data

    @property
    def subset_directional_variables(self) -> List[str]:
        return self._subset_directional_variables

    @property
    def target_directional_variables(self) -> List[str]:
        return self._target_directional_variables

    @property
    def subset_processed_variables(self) -> List[str]:
        return self._subset_processed_variables

    @property
    def target_processed_variables(self) -> List[str]:
        return self._target_processed_variables

    @property
    def subset_custom_scale_factor(self) -> dict:
        return self._subset_custom_scale_factor

    @property
    def target_custom_scale_factor(self) -> dict:
        return self._target_custom_scale_factor

    @property
    def subset_scale_factor(self) -> dict:
        return self._subset_scale_factor

    @property
    def target_scale_factor(self) -> dict:
        return self._target_scale_factor

    @property
    def rbf_coeffs(self) -> pd.DataFrame:
        return self._rbf_coeffs

    @property
    def opt_sigmas(self) -> dict:
        return self._opt_sigmas

    def _preprocess_subset_data(
        self, subset_data: pd.DataFrame, is_fit: bool = True
    ) -> pd.DataFrame:
        """
        This function preprocesses the subset data.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data to preprocess (could be a dataset to predict).
        is_fit : bool, optional
            Whether the data is being fit or not. Default is True.

        Returns
        -------
        pd.DataFrame
            The preprocessed subset data.

        Raises
        ------
        ValueError
            If the subset contains NaNs.

        Notes
        -----
        - This function preprocesses the subset data by:
            - Checking for NaNs.
            - Preprocessing directional variables.
            - Normalizing the data.
        """

        # Make copies to avoid modifying the original data
        subset_data = subset_data.copy()

        self.logger.info("Checking for NaNs in subset data")
        subset_data = self.check_nans(data=subset_data, raise_error=True)

        self.logger.info("Preprocessing subset data")
        for directional_variable in self.subset_directional_variables:
            var_u_component, var_y_component = self.get_uv_components(
                x_deg=subset_data[directional_variable].values
            )
            subset_data[f"{directional_variable}_u"] = var_u_component
            subset_data[f"{directional_variable}_v"] = var_y_component
            # Drop the original directional variable in subset_data
            subset_data.drop(columns=[directional_variable], inplace=True)
        self._subset_processed_variables = list(subset_data.columns)

        self.logger.info("Normalizing subset data")
        normalized_subset_data, subset_scale_factor = self.normalize(
            data=subset_data,
            custom_scale_factor=self.subset_custom_scale_factor
            if is_fit
            else self.subset_scale_factor,
        )

        self.logger.info("Subset data preprocessed successfully")

        if is_fit:
            self._subset_data = subset_data
            self._normalized_subset_data = normalized_subset_data
            self._subset_scale_factor = subset_scale_factor

        return normalized_subset_data.copy()

    def _preprocess_target_data(
        self,
        target_data: pd.DataFrame,
        normalize_target_data: bool = True,
    ) -> pd.DataFrame:
        """
        This function preprocesses the target data.

        Parameters
        ----------
        target_data : pd.DataFrame
            The target data to preprocess.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.

        Returns
        -------
        pd.DataFrame
            The preprocessed target data.

        Raises
        ------
        ValueError
            If the target contains NaNs.

        Notes
        -----
        - This function preprocesses the target data by:
            - Checking for NaNs.
            - Preprocessing directional variables.
            - Normalizing the data.
        """

        # Make copies to avoid modifying the original data
        target_data = target_data.copy()

        self.logger.info("Checking for NaNs in target data")
        target_data = self.check_nans(data=target_data, raise_error=True)

        self.logger.info("Preprocessing target data")
        for directional_variable in self.target_directional_variables:
            var_u_component, var_y_component = self.get_uv_components(
                x_deg=target_data[directional_variable].values
            )
            target_data[f"{directional_variable}_u"] = var_u_component
            target_data[f"{directional_variable}_v"] = var_y_component
            # Drop the original directional variable in target_data
            target_data.drop(columns=[directional_variable], inplace=True)
        self._target_processed_variables = list(target_data.columns)

        if normalize_target_data:
            self.logger.info("Normalizing target data")
            normalized_target_data, target_scale_factor = self.normalize(
                data=target_data,
                custom_scale_factor=self.target_custom_scale_factor,
            )
            self.is_target_normalized = True
            self._target_data = target_data.copy()
            self._normalized_target_data = normalized_target_data.copy()
            self._target_scale_factor = target_scale_factor.copy()
            self.logger.info("Target data preprocessed successfully")
            return normalized_target_data.copy()

        else:
            self.is_target_normalized = False
            self._target_data = target_data.copy()
            self._normalized_target_data = pd.DataFrame()
            self._target_scale_factor = {}
            self.logger.info("Target data preprocessed successfully")
            return target_data.copy()

    def _rbf_assemble(self, x, sigma):
        """
        Assemble the RBF matrix.

        Parameters
        ----------
        x : np.ndarray
            The data.
        sigma : float
            The sigma parameter for the kernel.

        Returns
        -------
        np.ndarray
            The data with all the calculated kernel values.
        """

        # Get the number of rows and columns in x
        dim, n = x.shape

        # Compute the pairwise distances
        dists = np.linalg.norm(x[:, :, np.newaxis] - x[:, np.newaxis, :], axis=0)

        # Apply the kernel function to the distances
        A = self.kernel_func(dists, sigma)

        # Subtract the smoothing parameter from the diagonal elements
        np.fill_diagonal(A, A.diagonal() - self.smooth)

        # Add the identity matrix to the matrix (polynomial term)
        P = np.hstack((np.ones((n, 1)), x.T))
        A = np.vstack(
            (np.hstack((A, P)), np.hstack((P.T, np.zeros((dim + 1, dim + 1)))))
        )

        return A

    def _calc_rbf_coeff(
        self, sigma: float, x: np.ndarray, y: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This function calculates the RBF coefficients for the given data.

        Parameters
        ----------
        sigma : float
            The sigma parameter for the kernel.
        x : np.ndarray
            The subset data used to interpolate.
        y : np.ndarray
            The target data to interpolate.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The RBF coefficients and the A matrix.
        """

        # Get the number of rows and columns in x
        m, n = x.shape

        # Assemble the A matrix
        A = self._rbf_assemble(x=x, sigma=sigma)

        # Concatenate y with zeros and reshape
        b = np.concatenate((y, np.zeros((m + 1,)))).reshape(-1, 1)

        # Calculate the RBF coefficients
        rbfcoeff, _, _, _ = np.linalg.lstsq(A, b, rcond=None)  # inverse

        return rbfcoeff, A

    def _cost_sigma(self, sigma: float, x: np.ndarray, y: np.ndarray) -> float:
        """
        This function is called by fminbound to minimize the cost function.

        Parameters
        ----------
        sigma : float
            The sigma parameter for the kernel.
        x : np.ndarray
            The subset data used to interpolate.
        y : np.ndarray
            The target data to interpolate.

        Returns
        -------
        float
            The cost value.
        """

        # Calculate RBF coefficients and A matrix
        rbf_coeff, A = self._calc_rbf_coeff(sigma=sigma, x=x, y=y)

        # Extract the top-left n x n submatrix from A
        m, n = x.shape
        A = A[:n, :n]

        # Compute the pseudo-inverse of the submatrix A
        invA = np.linalg.pinv(A)

        # Initialize residuals by subtracting the last m elements of rbf_coeff from y
        m1, n1 = rbf_coeff.shape
        kk = y - rbf_coeff[m1 - m - 1]

        # Adjust residuals by subtracting the product of rbf_coeff and x
        # DEPRECATED: The loop is replaced by the vectorized operation below
        # for i in range(m):
        #     kk = kk - rbf_coeff[m1 - m + i] * x[i, :]
        kk -= np.dot(rbf_coeff[m1 - m :].T, x).reshape(-1)

        # Calculate the cost by multiplying invA with kk and normalizing by the diagonal elements of invA
        ceps = np.dot(invA, kk) / np.diagonal(invA)

        # Return the norm of ceps, representing the cost
        yy = np.linalg.norm(ceps)

        return yy

    def _calc_opt_sigma(
        self,
        target_variable: np.ndarray,
        subset_variables: np.ndarray,
        iteratively_update_sigma: bool = False,
    ) -> float:
        """
        This function calculates the optimal sigma for the given target variable.

        Parameters
        ----------
        target_variable : np.ndarray
            The target variable to interpolate.
        subset_variables : np.ndarray
            The subset variables used to interpolate.
        iteratively_update_sigma : bool, optional
            Whether to iteratively update the sigma parameter. Default is False.

        Returns
        -------
        float
            The optimal sigma.
        """

        t0 = time.time()
        # Initialize sigma_min, sigma_max, and d_sigma
        sigma_min, sigma_max, d_sigma = self.sigma_min, self.sigma_max, 0

        if self.sigma_opt is not None:
            # Optimize sigma using the specified sigma_opt
            opt_sigma = fmin(
                func=self._cost_sigma,
                x0=self.sigma_opt,
                args=(subset_variables, target_variable),
                disp=0,
            )
            if iteratively_update_sigma:
                self._sigma_opt = opt_sigma
        else:
            # Loop until sigma_diff is less than the specified sigma_diff
            while d_sigma < self.sigma_diff:
                opt_sigma = fminbound(
                    func=self._cost_sigma,
                    x1=sigma_min,
                    x2=sigma_max,
                    args=(subset_variables, target_variable),
                    disp=0,
                )
                lm_min = np.abs(opt_sigma - sigma_min)
                lm_max = np.abs(opt_sigma - sigma_max)
                if lm_min < self.sigma_diff:
                    sigma_min = sigma_min - sigma_min / 2
                elif lm_max < self.sigma_min:
                    sigma_max = sigma_max + sigma_max / 2
                d_sigma = np.nanmin([lm_min, lm_max])
            if iteratively_update_sigma:
                self._sigma_opt = opt_sigma

        # Calculate the time taken to optimize sigma
        t1 = time.time()
        self.logger.info(f"Optimal sigma: {opt_sigma} - Time: {t1 - t0:.2f} seconds")

        # Calculate the RBF coefficients for the optimal sigma
        rbf_coeff, _ = self._calc_rbf_coeff(
            sigma=opt_sigma, x=subset_variables, y=target_variable
        )

        return rbf_coeff, opt_sigma

    def _rbf_variable_interpolation(
        self,
        normalized_dataset: pd.DataFrame,
        opt_sigma: float,
        rbf_coeff: np.ndarray,
        num_points_subset: int,
        num_vars_subset: int,
    ) -> np.ndarray:
        """
        Interpolates the surface for a variable.

        normalized_dataset : pd.DataFrame
            The normalized dataset.
        opt_sigma : float
            The optimal sigma calculated for variable.
        rbf_coeff : np.ndarray
            The fitted coefficients for variable.
        num_points_subset : int
            The number of points used in the fitting.
        num_vars_subset : int
            The number of variables used in the fitting.

        np.ndarray
            The interpolated variable.
        """

        r = np.linalg.norm(
            normalized_dataset.values[:, np.newaxis, :]
            - self.normalized_subset_data.values[np.newaxis, :, :],
            axis=2,
        )
        kernel_values = self.kernel_func(r, opt_sigma)
        linear_part = np.dot(
            normalized_dataset.values,
            rbf_coeff[
                num_points_subset + 1 : num_points_subset + 1 + num_vars_subset
            ].T,
        )

        return (
            rbf_coeff[num_points_subset]
            + np.dot(kernel_values, rbf_coeff[:num_points_subset])
            + linear_part
        )

    def _rbf_interpolate(
        self, dataset: pd.DataFrame, num_threads: int = None
    ) -> pd.DataFrame:
        """
        This function interpolates the dataset.

        Parameters
        ----------
        dataset : pd.DataFrame
            The dataset to interpolate (must have same variables as subset).
        num_threads : int, optional
            The number of threads to use for the interpolation. Default is None.

        Returns
        -------
        pd.DataFrame
            The interpolated dataset (with all target variables).
        """

        normalized_dataset = self._preprocess_subset_data(
            subset_data=dataset, is_fit=False
        )

        # Get the number of rows and columns in subset and dataset
        num_vars_subset, num_points_subset = self.normalized_subset_data.T.shape
        _, num_points_dataset = normalized_dataset.T.shape

        # Initialize the interpolated dataset
        interpolated_array = np.zeros(
            (num_points_dataset, len(self.target_processed_variables))
        )

        # Loop through the target variables
        if num_threads is not None:
            # self.set_num_processors_to_use(num_processors=num_threads)
            num_threads = min(num_threads, self.get_num_processors_available())
            self.logger.info(f"Using {num_threads} threads for interpolation.")
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                rbf_variable_calculation = {
                    executor.submit(
                        self._rbf_variable_interpolation,
                        normalized_dataset,
                        self._opt_sigmas[target_var],
                        self._rbf_coeffs[target_var].values,
                        num_points_subset,
                        num_vars_subset,
                    ): (i_var, target_var)
                    for i_var, target_var in enumerate(self.target_processed_variables)
                }
                for future in as_completed(rbf_variable_calculation):
                    i_rbf_var, rbf_variable = rbf_variable_calculation[future]
                    try:
                        interpolated_var = future.result()
                        interpolated_array[:, i_rbf_var] = interpolated_array
                    except Exception as exc:
                        self.logger.error(
                            f"Job for {rbf_variable} generated an exception: {exc}."
                        )
        else:
            for i_var, target_var in enumerate(self.target_processed_variables):
                self.logger.info(f"Interpolating target variable {target_var}")
                interpolated_var = self._rbf_variable_interpolation(
                    normalized_dataset=normalized_dataset,
                    opt_sigma=self._opt_sigmas[target_var],
                    rbf_coeff=self._rbf_coeffs[target_var].values,
                    num_points_subset=num_points_subset,
                    num_vars_subset=num_vars_subset,
                )
                interpolated_array[:, i_var] = interpolated_var

        return pd.DataFrame(interpolated_array, columns=self.target_processed_variables)

    @validate_data_rbf
    def fit(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
        subset_directional_variables: List[str] = [],
        target_directional_variables: List[str] = [],
        subset_custom_scale_factor: dict = {},
        normalize_target_data: bool = True,
        target_custom_scale_factor: dict = {},
        num_threads: int = None,
        iteratively_update_sigma: bool = False,
    ) -> None:
        """
        Fits the model to the data.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data used to fit the model.
        target_data : pd.DataFrame
            The target data used to fit the model.
        subset_directional_variables : List[str], optional
            The subset directional variables. Default is [].
        target_directional_variables : List[str], optional
            The target directional variables. Default is [].
        subset_custom_scale_factor : dict, optional
            The custom scale factor for the subset data. Default is {}.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.
        target_custom_scale_factor : dict, optional
            The custom scale factor for the target data. Default is {}.
        num_threads : int, optional
            The number of threads to use for the optimization. Default is None.
        iteratively_update_sigma : bool, optional
            Whether to iteratively update the sigma parameter. Default is False.

        Notes
        -----
        - This function fits the RBF model to the data by:
            1. Preprocessing the subset and target data.
            2. Calculating the optimal sigma for the target variables.
            3. Storing the RBF coefficients and optimal sigmas.
        - The number of threads to use for the optimization can be specified.
        """

        self._subset_directional_variables = subset_directional_variables
        self._target_directional_variables = target_directional_variables
        self._subset_custom_scale_factor = subset_custom_scale_factor
        self._target_custom_scale_factor = target_custom_scale_factor
        subset_data = self._preprocess_subset_data(subset_data=subset_data)
        target_data = self._preprocess_target_data(
            target_data=target_data,
            normalize_target_data=normalize_target_data,
        )

        self.logger.info("Fitting RBF model to the data")
        # RBF fitting for all variables
        rbf_coeffs, opt_sigmas = {}, {}

        if num_threads is not None:
            # self.set_num_processors_to_use(num_processors=num_threads)
            num_threads = min(num_threads, self.get_num_processors_available())
            self.logger.info(f"Using {num_threads} threads for optimization.")
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                rbf_variable_calculation = {
                    executor.submit(
                        self._calc_opt_sigma,
                        target_data[target_var].values,
                        subset_data.values.T,
                        iteratively_update_sigma,
                    ): target_var
                    for target_var in target_data.columns
                }
                for future in as_completed(rbf_variable_calculation):
                    rbf_variable = rbf_variable_calculation[future]
                    try:
                        rbf_coeff, opt_sigma = future.result()
                        rbf_coeffs[rbf_variable] = rbf_coeff.flatten()
                        opt_sigmas[rbf_variable] = opt_sigma
                    except Exception as exc:
                        self.logger.error(
                            f"Job for {rbf_variable} generated an exception: {exc}."
                        )
        else:
            for target_var in target_data.columns:
                self.logger.info(f"Fitting RBF for variable {target_var}")
                target_var_values = target_data[target_var].values
                rbf_coeff, opt_sigma = self._calc_opt_sigma(
                    target_variable=target_var_values,
                    subset_variables=subset_data.values.T,
                    iteratively_update_sigma=iteratively_update_sigma,
                )
                rbf_coeffs[target_var] = rbf_coeff.flatten()
                opt_sigmas[target_var] = opt_sigma

        # Store the RBF coefficients and optimal sigmas
        self._rbf_coeffs = pd.DataFrame(rbf_coeffs)
        self._opt_sigmas = opt_sigmas

        # Set the is_fitted attribute to True
        self.is_fitted = True

    def predict(self, dataset: pd.DataFrame, num_threads: int = None) -> pd.DataFrame:
        """
        Predicts the data for the provided dataset.

        Parameters
        ----------
        dataset : pd.DataFrame
            The dataset to predict (must have same variables than subset).
        num_threads : int, optional
            The number of threads to use for the interpolation. Default is None.

        Returns
        -------
        pd.DataFrame
            The interpolated dataset.

        Raises
        ------
        ValueError
            If the model is not fitted.

        Notes
        -----
        - This function predicts the data by:
            1. Reconstructing the data using the fitted coefficients.
            2. Denormalizing the target data if normalize_target_data is True.
            3. Calculating the degrees for the target directional variables.
        """

        if self.is_fitted is False:
            raise RBFError("RBF model must be fitted before predicting.")

        self.logger.info("Reconstructing data using fitted coefficients.")
        interpolated_target = self._rbf_interpolate(
            dataset=dataset, num_threads=num_threads
        )
        if self.is_target_normalized:
            self.logger.info("Denormalizing target data")
            interpolated_target = self.denormalize(
                normalized_data=interpolated_target,
                scale_factor=self.target_scale_factor,
            )
        for directional_variable in self.target_directional_variables:
            self.logger.info(f"Calculating target degrees for {directional_variable}")
            interpolated_target[directional_variable] = self.get_degrees_from_uv(
                xu=interpolated_target[f"{directional_variable}_u"].values,
                xv=interpolated_target[f"{directional_variable}_v"].values,
            )

        return interpolated_target

    def fit_predict(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
        dataset: pd.DataFrame,
        subset_directional_variables: List[str] = [],
        target_directional_variables: List[str] = [],
        subset_custom_scale_factor: dict = {},
        normalize_target_data: bool = True,
        target_custom_scale_factor: dict = {},
        num_threads: int = None,
        iteratively_update_sigma: bool = False,
    ) -> pd.DataFrame:
        """
        Fits the model to the subset and predicts the interpolated dataset.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data used to fit the model.
        target_data : pd.DataFrame
            The target data used to fit the model.
        dataset : pd.DataFrame
            The dataset to predict (must have same variables than subset).
        subset_directional_variables : List[str], optional
            The subset directional variables. Default is [].
        target_directional_variables : List[str], optional
            The target directional variables. Default is [].
        subset_custom_scale_factor : dict, optional
            The custom scale factor for the subset data. Default is {}.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.
        target_custom_scale_factor : dict, optional
            The custom scale factor for the target data. Default is {}.
        num_threads : int, optional
            The number of threads to use for the optimization. Default is None.
        iteratively_update_sigma : bool, optional
            Whether to iteratively update the sigma parameter. Default is False.

        Returns
        -------
        pd.DataFrame
            The interpolated dataset.

        Notes
        -----
        - This function fits the model to the subset and predicts the interpolated dataset.
        """

        self.fit(
            subset_data=subset_data,
            target_data=target_data,
            subset_directional_variables=subset_directional_variables,
            target_directional_variables=target_directional_variables,
            subset_custom_scale_factor=subset_custom_scale_factor,
            normalize_target_data=normalize_target_data,
            target_custom_scale_factor=target_custom_scale_factor,
            num_threads=num_threads,
            iteratively_update_sigma=iteratively_update_sigma,
        )

        return self.predict(dataset=dataset, num_threads=num_threads)



import numpy as np


def energy_spectrum(hs, tp, gamma, duration):
    """
    spec: Dataset with vars for the whole partitions
    S(f,dir) = S(f) * D(dir)
    S(f) ----- D(dir)
    Meshgrid de x freqs - dir - z energy
    """

    # Defining frequency series - tend length
    freqs = np.linspace(0.02, 1, duration)

    S = []
    fp = 1 / tp

    for f in freqs:
        if f <= fp:
            sigma = 0.07
        if f > fp:
            sigma = 0.09

        Beta = (0.06238 / (0.23 + 0.0336 * gamma - 0.185 * (1.9 + gamma) ** -1)) * (
            1.094 - 0.01915 * np.log(gamma)
        )
        Sf = (
            Beta
            * (hs**2)
            * (tp**-4)
            * (f**-5)
            * np.exp(-1.25 * (tp * f) ** -4)
            * gamma ** (np.exp((-((tp * f - 1) ** 2)) / (2 * sigma**2)))
        )
        S.append(Sf)

    return S


def series_Jonswap(waves):
    """
    Generate surface elevation from PSD df = 1/tendc

    waves - dictionary
              T       - Period (s)
              H       - Height (m)
              gamma   - Jonswap spectrum  peak parammeter
              warmup  - spin up time (s)
              deltat  - delta time (s)
              tendc   - simulation period (s)

    returns 2D numpy array with series time and elevation
    """

    # waves properties
    hs = waves["H"]
    tp = waves["T"]
    gamma = waves["gamma"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    # why? wee user manual
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum
    S = energy_spectrum(hs, tp, gamma, duration)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_TMA(waves, depth):
    """
    Generate surface elevation from PSD df = 1/tendc

    waves - dictionary
              T       - Period (s)
              H       - Height (m)
              gamma   - Jonswap spectrum  peak parammeter
              warmup  - spin up time (s)
              deltat  - delta time (s)
              tendc   - simulation period (s)

    returns 2D numpy array with series time and elevation
    """

    # waves properties
    hs = waves["H"]
    tp = waves["T"]
    gamma = waves["gamma"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]
    g = 9.801

    # series duration
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    # why? wee user manual
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate JONSWAP energy spectrum
    S = energy_spectrum(hs, tp, gamma, duration)
    m0_Jonswap = np.trapz(S, x=freqs)

    # Transform JONSWAP into TMA by (omega, h)
    S_TMA = []
    for pf, f in enumerate(freqs):
        """
        omega = 2 * np.pi * f
        
        thv = omega * np.sqrt(depth / g)
        if thv < 1: 
            fi = 0.5 * (thv)**2
        elif (thv >= 1) and (thv < 2):
            fi = 1 - 0.5 * (2 - thv)**2
        else:
            fi = 1
        """
        # S_TMA.append(S[pf] * fi)

        # As in Holthuijsen
        L = waves_dispersion(1 / f, depth)[0]
        k = 2 * np.pi / L
        n = 0.5 * (1 + ((2 * k * depth) / (np.sinh(2 * k * depth))))
        fi = (1 / (2 * n)) * np.tanh(k * depth)

        S_TMA.append(S[pf] * fi)

    m0_TMA = np.trapz(S_TMA, x=freqs)

    S_TMA_mod = [i * (m0_Jonswap / m0_TMA) for i in S_TMA]

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S_TMA_mod[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_TMA_bimodal(waves, depth):
    """
    Generate surface elevation from PSD df = 1/tendc

    waves - dictionary
              T       - Period (s)
              H       - Height (m)
              gamma   - Jonswap spectrum  peak parammeter
              warmup  - spin up time (s)
              deltat  - delta time (s)
              tendc   - simulation period (s)

    returns 2D numpy array with series time and elevation
    """

    # waves properties
    hs1 = waves["Hs1"]
    tp1 = waves["Tp1"]
    gamma1 = waves["gamma1"]
    hs2 = waves["Hs2"]
    tp2 = waves["Tp2"]
    gamma2 = waves["gamma2"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    # TODO: puede que haya algun problema en esta suma
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum for each wave system
    S1 = energy_spectrum(hs1, tp1, gamma1, duration)
    S2 = energy_spectrum(hs2, tp2, gamma2, duration)

    m01_Jonswap = np.trapz(S1, x=freqs)
    m02_Jonswap = np.trapz(S2, x=freqs)

    # Transform JONSWAP into TMA by (omega, h)
    S_TMA_1, S_TMA_2 = [], []
    for pf, f in enumerate(freqs):
        # As in Holthuijsen
        L = waves_dispersion(1 / f, depth)[0]
        k = 2 * np.pi / L
        n = 0.5 * (1 + ((2 * k * depth) / (np.sinh(2 * k * depth))))
        fi = (1 / (2 * n)) * np.tanh(k * depth)

        S_TMA_1.append(S1[pf] * fi)
        S_TMA_2.append(S2[pf] * fi)

    m0_TMA_1 = np.trapz(S_TMA_1, x=freqs)
    m0_TMA_2 = np.trapz(S_TMA_2, x=freqs)

    S_TMA_mod_1 = [i * (m01_Jonswap / m0_TMA_1) for i in S_TMA_1]
    S_TMA_mod_2 = [i * (m02_Jonswap / m0_TMA_2) for i in S_TMA_2]

    # bimodal wave spectra
    S = np.sum([S_TMA_mod_1, S_TMA_mod_2], axis=0)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_Jonswap_bimodal(waves):
    """
    Generate surface elevation from PSD df = 1/tendc

    waves - dictionary
              T       - Period (s)
              H       - Height (m)
              gamma   - Jonswap spectrum  peak parammeter
              warmup  - spin up time (s)
              deltat  - delta time (s)
              tendc   - simulation period (s)

    returns 2D numpy array with series time and elevation
    """

    # waves properties
    hs1 = waves["Hs1"]
    tp1 = waves["Tp1"]
    gamma1 = waves["gamma1"]
    hs2 = waves["Hs2"]
    tp2 = waves["Tp2"]
    gamma2 = waves["gamma2"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    # TODO: puede que haya algun problema en esta suma
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum for each wave system
    S1 = energy_spectrum(hs1, tp1, gamma1, duration)
    S2 = energy_spectrum(hs2, tp2, gamma2, duration)
    # bimodal wave spectra
    S = np.sum([S1, S2], axis=0)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_regular_monochromatic(waves):
    """
    Generates monochromatic regular waves series

    waves - dictionary
              T      - Period (s)
              H      - Height (m)
              WL     - Water level (m)
              warmup - spin up time (s)
              deltat - delta time (s)
              tendc  - simulation period (s)

    returns 2D numpy array with series time and elevation
    """

    # waves properties
    T = waves["T"]
    H = waves["H"]
    WL = waves["WL"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    duration = tendc + int(warmup)
    time = np.arange(0, duration, deltat)

    # series elevation
    teta = (H / 2) * np.cos((2 * np.pi / T) * time)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_regular_bichromatic(waves):
    """
    Generates bichromatic regular waves series

    waves - dictionary
              T1     - Period component 1 (s)
              T2     - Period component 2 (s)
              H      - Height (m)
              WL     - Water level (m)
              warmup - spin up time (s)
              deltat - delta time (s)
              tendc  - simulation period (s)
    """

    # waves properties
    T1 = waves["T1"]
    T2 = waves["T2"]
    H = waves["H"]
    WL = waves["WL"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    duration = tendc + int(warmup)
    time = np.arange(0, duration, deltat)

    # series elevation
    teta = (H / 2) * np.cos((2 * np.pi / T1) * time) + (H / 2) * np.cos(
        (2 * np.pi / T2) * time
    )

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def waves_dispersion(T, h):
    "Solve the wave dispersion relation"

    L1 = 1
    L2 = ((9.81 * T**2) / 2 * np.pi) * np.tanh(h * 2 * np.pi / L1)
    umbral = 1

    while umbral > 0.1:
        L2 = ((9.81 * T**2) / (2 * np.pi)) * np.tanh((h * 2 * np.pi) / L1)
        umbral = np.abs(L2 - L1)
        L1 = L2

    L = L2
    k = (2 * np.pi) / L
    c = np.sqrt(9.8 * np.tanh(k * h) / k)

    return (L, k, c)



import os
import os.path as op
import glob
from natsort import natsorted

import xarray as xr
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mcolors

import cartopy.crs as ccrs
import cartopy
import cartopy.feature as cfeature

from olas import estela

def preprocess_monthly_files(ds):
    filename = os.path.basename(ds.encoding['source'])
    year_month = filename.split('_')[1][:6]
    timenew = np.array(xr.DataArray(pd.to_datetime(year_month, format='%Y%m')))
    ds['time'] = (('time'), [timenew])
    return ds

def preprocess_monthly_files_sea(ds):
    filename = os.path.basename(ds.encoding['source'])
    year_month = filename.split('_')[2][:6]
    timenew = np.array(xr.DataArray(pd.to_datetime(year_month, format='%Y%m')))
    ds['time'] = (('time'), [timenew])
    return ds

plot_defaults = {
        'extent' : None,
        'figsize' : None, #[20, 8],
        'vmin': 0,
        'vmin_anomaly': None,
        'vmax': None,
        'vmax_anomaly': None,
        'central_longitude' : 180,
        'cbar': True,
        'marker': '.',
        'fontsize' : 12,
        'cmap' : 'rainbow',
        'cmap_anomaly' : 'RdBu_r',
        'c_color': 'plum',
        'alpha':1
    }

    
class ESTELA:

    '''
    This class implements the calculation of ESTELA using the python that can be found at: https://github.com/jorgeperezg/olas, which
    is based on the following publication [Pérez et al., 2014](https://doi.org/10.1007/s10236-014-0740-7)
    
    '''

    def __init__(self, p_waves, save_monthly = False, only_sea = False):
        
        self.p_waves = p_waves
        self.save_monthly = save_monthly
        self.only_sea = only_sea
        self.site = None
        self.lon0 = None
        self.lat0 = None
        self.PATHS = None

        #These wave parameter names are for CSIRO partitions. If other database is used, these should be changed.
        self.waves_parsm = vars = ['MAPSTA', 'hs0', 'hs1', 'hs2', 'hs3',
                                            'si0', 'si1', 'si2', 'si3',
                                            'th0', 'th1', 'th2', 'th3',
                                            'tp0', 'tp1', 'tp2', 'tp3']

        if save_monthly:
            self.groupers = ["time.month"]
        else:
            self.groupers = ["ALL", "time.season",  "time.month"] 

    def set_site(self, site, location):

        self.site = site
        self.location = location
        self.lon0 = location[0]
        self.lat0 = location[1]

        if location[0] < 0:
            raise ESTELAError('Location longitude must be within [0-360]. Location given: {0}'.format(location))

    def set_years(self, years):

        p_waves = self.p_waves
        y_start = years[0]
        y_end = years[1]
        
        self.years = years
        self.y_start = y_start
        self.y_end = y_end

        # database
        datafiles = natsorted(glob.glob(op.join(p_waves, "*.nc")))
        #Define range of years for the analysis
        years = np.arange(y_start, y_end+1)
        #select paths in those years
        PATHS = []
        for s in datafiles:
            for y in years:
                if y.astype('str') in s.split('.nc')[0].split('.')[-1][:-2]:
                    PATHS.append(s)
        
        self.PATHS = natsorted(PATHS)


    
    def get_estela(self, nblocks = 1, do_estela = True):

        if not (self.site is not None or self.lon0 is not None or self.lat0 is not None or self.PATHS is not None):
            raise ESTELAError('Site and Years need to be defined before using the function set_location and set_years')
            
    
        PATHS = self.PATHS
        lon0 = self.lon0
        lat0 = self.lat0
        save_monthly = self.save_monthly
        groupers = self.groupers
        site = self.site
        only_sea = self.only_sea
    
        # Generate paths for saving
    
        p_site = op.join(os.getcwd(), site)
        self.p_site = p_site
        # make output folder
        if not op.isdir(p_site):
            os.makedirs(p_site)

        if only_sea:
            p_estela_o = op.join(p_site, 'estela_sea.nc')
            p_estela_monthly = op.join(p_site, 'monthly_estela_sea.nc')
            name_estela = 'estela_sea'
        else:
            p_estela_o = op.join(p_site, 'estela.nc')
            p_estela_monthly = op.join(p_site, 'monthly_estela.nc')
            name_estela = 'estela'
            
            
        if do_estela: 
            # Calculate estela
            if save_monthly:
                for i in range(len(PATHS)):
                    p_estela = op.join(p_site, '{}_{}.nc'.format(name_estela, PATHS[i].split('/')[-1].split('.')[-2]))
                    
                    p_file_estela = PATHS[i].split('/')[-1].split('.')[-2]
                    y = np.double(p_file_estela[:-2])
                    m = np.double(p_file_estela[-2:])

                    if not op.isfile(p_estela):
                        if only_sea:
                            if ((y ==2013) and (m>=6)) or (y>2013):
                                est = estela.calc(PATHS[i], lat0, lon0, "phs0", "ptp0", "pdir0", "pspr0", groupers=groupers, nblocks=nblocks)
                            else:
                                est = estela.calc(PATHS[i], lat0, lon0, "hs0", "tp0", "th0", "si0", groupers=groupers, nblocks=nblocks)
                        else:
                            if ((y ==2013) and (m>=6)) or (y>2013):
                                est = estela.calc(PATHS[i], lat0, lon0, "phs.", "ptp.", "pdir.", "pspr.", groupers=groupers, nblocks=nblocks)
                            else:
                                est = estela.calc(PATHS[i], lat0, lon0, "hs.", "tp.", "th.", "si.", groupers=groupers, nblocks=nblocks)
                        est.to_netcdf(path = p_estela)

                p_estelas = op.join(p_site, '{}_*.nc'.format(name_estela))
                if only_sea:
                    est = xr.open_mfdataset(p_estelas, preprocess=preprocess_monthly_files_sea)
                else:
                    est = xr.open_mfdataset(p_estelas, preprocess=preprocess_monthly_files)
                est = est.assign({'estela_mask':(('latitude','longitude'), np.where(est.F.isel(time = 0).values > 0.001, 1, np.nan))})
                #est = est.mean(dim = 'time')
                est = est.sortby(est.latitude, ascending = False)
                est.to_netcdf(p_estela_monthly)

                self.p_estela_monthly = p_estela_monthly

                est = est.mean(dim = 'time')
                est.to_netcdf(p_estela_o)
                self.path_estela = p_estela_o
                
            else:
                if not op.isfile(p_estela_o):
                    #est = estela.calc(PATHS, lat0, lon0, "hs.", "tp.", "th.", 20, groupers=groupers, nblocks=nblocks)
                    est = estela.calc(PATHS, lat0, lon0, "hs.", "tp.", "th.", "si.", groupers=groupers, nblocks=nblocks)
                    
                    est = est.assign({'estela_mask':(('latitude','longitude'), np.where(est.F.isel(time = 0).values > 0.001, 1, np.nan))})
                    est = est.sortby(est.latitude, ascending = False)
                    est.to_netcdf(path = p_estela_o)
            
            est = xr.open_dataset(p_estela_o, decode_times = False)
            os.remove(p_estela_o)
            est['traveltime'] = est['traveltime']/3600/10**9/24
            est.to_netcdf(path = p_estela_o)

        else:

            if not op.isfile(p_estela_o):
                raise ESTELAError('Estela file not found at {0}. This could happen because do_estela parameter is set to False, and estela has not been run yet. Consider changing do_estela = True'.format(p_estela_o))
                
            # Load previously calculated estela
            est = xr.open_dataset(p_estela_o, decode_times = False)

        self.estela = est

    def load_estela_monthly(self):

        estela_monthly = xr.open_dataset(self.p_estela_monthly)
        estela_monthly['traveltime'] = estela_monthly['traveltime']/3600/10**9/24

        return estela_monthly

    def load_estela_monthly_mean(self):

        only_sea = self.only_sea

        if only_sea:
            p_estela_monthly_mean = op.join(self.p_site, 'monthly_mean_estela_sea.nc')
        else:
            p_estela_monthly_mean = op.join(self.p_site, 'monthly_mean_estela.nc')
            
        est = self.load_estela_monthly()
        est = est.groupby(est.time.dt.month).mean()
        est['month_name'] = (('month'), ['January', 'February', 'March', 'April', 'May', 'June', 'July',
                            'August', 'September', 'October', 'November', 'December'])
        if os.path.exists(p_estela_monthly_mean): os.remove(p_estela_monthly_mean)
        est.to_netcdf(p_estela_monthly_mean)

        return xr.open_dataset(p_estela_monthly_mean, decode_times = False)

    def load_estela_seasonal(self):

        only_sea = self.only_sea

        if only_sea:
             p_estela_seasonal = op.join(self.p_site, 'seasonal_estela_sea.nc')
        else:
            p_estela_seasonal = op.join(self.p_site, 'seasonal_estela.nc')
        
        estela_monthly = xr.open_dataset(self.p_estela_monthly)
        # Define quarters. DJF MAM JJA SON
        estela_monthly['season'] = (('time'), np.digitize(estela_monthly['time.month'], [3, 6, 9, 12]) % 4)
        
        # Group data
        estela_seasonal = estela_monthly.groupby('season').mean()
        estela_seasonal['season_name'] = (('season'), np.array(['DJF', 'MAM', 'JJA', 'SON']))
        estela_seasonal['traveltime'] = estela_seasonal['traveltime']/3600/10**9/24

        if os.path.exists(p_estela_seasonal): os.remove(p_estela_seasonal)
        estela_seasonal.to_netcdf(p_estela_seasonal)

        return xr.open_dataset(p_estela_seasonal, decode_times = False)

    ##############################################
    ################## Plotting ##################
    ##############################################

    
    def plot_map_features(self, ax, land_color = cfeature.COLORS['land']):
        
        
        #ax.add_feature(cfeature.BORDERS, linestyle=':')
        #ax.add_feature(cfeature.LAND, edgecolor=None, color = land_color, zorder = 1)
        ax.add_feature(cfeature.OCEAN, color='lightblue', alpha = .5)
        #ax.add_feature(cfeature.COASTLINE, linewidth=1, edgecolor='black',  zorder = 2)
        land_10m = cartopy.feature.NaturalEarthFeature(
                            'physical', 'land', '10m',
                            edgecolor = 'black',
                            linewidth = .5,
                            facecolor = cfeature.COLORS['land'],
                            zorder = 1,
                                )
        ax.add_feature(land_10m)
        ax.gridlines()

    def get_extent(self, est, plot_params):
        
        if not plot_params['extent']:
            lon0, lon1 = np.nanmin(est.longitude[np.where(est.estela_mask == 1)[1]]), np.nanmax(est.longitude[np.where(est.estela_mask == 1)[1]])
            lat0, lat1 = np.nanmin(est.latitude[np.where(est.estela_mask == 1)[0]]), np.nanmax(est.latitude[np.where(est.estela_mask == 1)[0]])
            extent = (lon0, lon1, lat0, lat1)
            plot_params['extent'] = extent
        else:
            extent = plot_params['extent']

        return extent

    def get_figsize(self, plot_params, N = 15, ratio = 1):

        if not plot_params['figsize']:
            figsize = [N, N/ratio]
        else:
            figsize = plot_params['figsize']

        return figsize
    
    def plot_estela(self, est, plot_params, anomaly = False, fig = [], ax = []):

        if anomaly:
            mean_estela = self.estela
            energy_mean = mean_estela.F.values #* mean_estela.estela_mask.values

        extent = self.get_extent(est, plot_params)  
        ratio = ((extent[1]-extent[0])/(extent[3]-extent[2]))

     # generate figure and axes
        if not fig:

            figsize = self.get_figsize(plot_params, N = 15, ratio = ratio)
 
            fig = plt.figure(figsize = figsize)
            
            ax = plt.axes(projection = ccrs.PlateCarree(central_longitude = plot_params['central_longitude']))

            
        # cartopy land feature
        self.plot_map_features(ax)
    
        # plot estela
        energy = est.F.values #* est.estela_mask.values
        if anomaly:
            energy -= energy_mean

        if anomaly:
            vmin, vmax = plot_params['vmin_anomaly'], plot_params['vmax_anomaly']
            cmap_energy = plot_params['cmap_anomaly']
        else:
            vmin, vmax = plot_params['vmin'], plot_params['vmax']
            cmap_energy = plot_params['cmap']
            
        p1 = ax.contourf(
            est.longitude.values, est.latitude.values, energy, 100,
            transform = ccrs.PlateCarree(),
            zorder = 2, alpha = plot_params['alpha'],
            vmin = vmin, vmax = vmax,
            cmap = cmap_energy,
        )

        if not anomaly:
            cim = ax.contour(
                est.longitude.values, est.latitude.values, est.traveltime.values * est.estela_mask.values,
                levels = np.arange(1, 26, 1),
                transform = ccrs.PlateCarree(),
                zorder = 3,
                linestyles = ':',
                colors = plot_params['c_color'],
            )
            plt.clabel(cim, inline = True, fontsize = plot_defaults['fontsize'])

        ax.plot(self.lon0, self.lat0, transform = ccrs.PlateCarree(), marker = '*', color = 'black', markersize = 12, zorder = 5)
    
        if plot_params['cbar']:
            plt.colorbar(p1, shrink = .7).set_label('Energy', fontsize = plot_params['fontsize'])
    
        # set extent
        if plot_params['extent']:
            ax.set_extent(plot_params['extent'], crs = ccrs.PlateCarree())

        return p1

    def plot_mean_estela(self, fig = [], ax = [], custom_params = False):
        
        '''
        Plots estela map
    
        self.estela    - (xarray.Dataset) estela dataset [traveltime and F variables]
        '''

        est = self.estela
        
        plot_params = {**plot_defaults, **custom_params} if custom_params else plot_defaults  
        
        self.plot_estela(est, plot_params)
    
    def plot_season_estela(self, fig = [], ax = [], anomaly = False, custom_params = {'cbar' : False}):
        
        '''
        Plots estela map
    
        self.estela    - (xarray.Dataset) estela dataset [traveltime and F variables]
        '''

        est = self.load_estela_seasonal()

        en = est.F - self.estela.F
        lim = np.nanmax([np.abs(np.nanmin(en)), np.abs(np.nanmax(en))])
        plot_defaults['vmin_anomaly'] = -lim
        plot_defaults['vmax_anomaly'] = lim

        plot_defaults['vmax'] = est.F.max()
        
        plot_params = {**plot_defaults, **custom_params} if custom_params else plot_defaults

        extent = self.get_extent(est, plot_params)  
        ratio = ((extent[1]-extent[0])/(extent[3]-extent[2]))

        figsize = self.get_figsize(plot_params, N = 18, ratio = ratio)

        fig, axs = plt.subplots(2, 2, figsize=figsize, 
                                subplot_kw={'projection': ccrs.PlateCarree(central_longitude = plot_params['central_longitude'])})
        
        for i, ax in enumerate(axs.flat):

            ax.set_title('{}'.format(est.isel(season=i).season_name.values), fontsize = plot_params['fontsize'])  # Título del subplot
            im = self.plot_estela(est.isel(season = i), plot_params, fig = fig, ax = ax, anomaly = anomaly)

        if anomaly:
            norm = mcolors.Normalize(vmin=plot_params['vmin_anomaly'], vmax=plot_params['vmax_anomaly'])
            cmap = plot_params['cmap_anomaly']
        else:
            norm = mcolors.Normalize(vmin=plot_params['vmin'], vmax=plot_params['vmax'])
            cmap = plot_params['cmap']

        cbar = plt.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), ax = axs, orientation='vertical', shrink = .7)
        cbar.set_label('Energy', fontsize = plot_params['fontsize'])


    def plot_monthly_estela(self, fig = [], ax = [], anomaly = False, custom_params = {'cbar' : False}):
        
        '''
        Plots estela map
    
        self.estela    - (xarray.Dataset) estela dataset [traveltime and F variables]
        '''

        est = self.load_estela_monthly_mean()
        
        en = est.F - self.estela.F
        lim = np.nanmax([np.abs(np.nanmin(en)), np.abs(np.nanmax(en))])
        plot_defaults['vmin_anomaly'] = -lim
        plot_defaults['vmax_anomaly'] = lim

        plot_defaults['vmax'] = est.F.max()
        
        plot_params = {**plot_defaults, **custom_params} if custom_params else plot_defaults

        extent = self.get_extent(est, plot_params)  
        ratio = ((extent[1]-extent[0])/(extent[3]-extent[2]))

        figsize = self.get_figsize(plot_params, N = 22, ratio = ratio*3/4)

        fig, axs = plt.subplots(4, 3, figsize=figsize, 
                                subplot_kw={'projection': ccrs.PlateCarree(central_longitude = plot_params['central_longitude'])})

        for i, ax in enumerate(axs.flat):

            ax.set_title('{}'.format(est.isel(month=i).month_name.values), fontsize = plot_params['fontsize'])
            im = self.plot_estela(est.isel(month = i), plot_params, fig = fig, ax = ax, anomaly = anomaly)


        if anomaly:
            norm = mcolors.Normalize(vmin=plot_params['vmin_anomaly'], vmax=plot_params['vmax_anomaly'])
            cmap = plot_params['cmap_anomaly']
        else:
            norm = mcolors.Normalize(vmin=plot_params['vmin'], vmax=plot_params['vmax'])
            cmap = plot_params['cmap']

        cbar = plt.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), ax = axs, orientation='vertical', shrink = .7)
        cbar.set_label('Energy', fontsize = plot_params['fontsize'])



class ESTELAError(Exception):
    """Custom exception for ESTELA class."""
    def __init__(self, message="ESTELA error occurred."):
        self.message = message
        super().__init__(self.message)
        


import numpy as np
import xarray as xr


def generate_grid_parameters(bathy_data: xr.DataArray) -> dict:
    """
    Generate the grid parameters for the SWAN model.

    Parameters
    ----------
    bathy_data : xr.DataArray
        Bathymetry data.
        Must have the following dimensions:
        - lon: longitude
        - lat: latitude

    Returns
    -------
    dict
        Grid parameters for the SWAN model.

    Contact @bellidog on GitHub for more information.
    """

    return {
        "xpc": np.nanmin(bathy_data.lon),  # x origin
        "ypc": np.nanmin(bathy_data.lat),  # y origin
        "alpc": 0,  # x-axis direction
        "xlenc": np.nanmax(bathy_data.lon)
        - np.nanmin(bathy_data.lon),  # grid length in x
        "ylenc": np.nanmax(bathy_data.lat)
        - np.nanmin(bathy_data.lat),  # grid length in y
        "mxc": len(bathy_data.lon) - 1,  # number mesh x, una menos pq si no SWAN peta
        "myc": len(bathy_data.lat) - 1,  # number mesh y, una menos pq si no SWAN peta
        "dxinp": bathy_data.lon[1].values
        - bathy_data.lon[0].values,  # size mesh x (resolution in x)
        "dyinp": bathy_data.lat[1].values
        - bathy_data.lat[0].values,  # size mesh y (resolution in y)
    }



import numpy as np
import xarray as xr
import keras.utils


class DataGenerator(keras.utils.Sequence):
    def __init__(
        self,
        msl_path,
        tp_path,
        num_images,
        sequential=False,
        batch_size=1,
    ):
        # create memory-mapped files for high_res and low_res datasets

        # inputs

        self.msl = xr.open_dataarray(msl_path).values[:, :64, :64]

        # outputs

        self.tp = xr.open_dataarray(tp_path).values[:, :64, :64]

        # set boolean for sequential or random dataset
        self.sequential = sequential
        # counter for keeping track of seuquential generator
        self.counter = 0
        self.num_images = num_images
        # set sequence len
        # flag for diffusion/unet
        self.batch_size = batch_size
        self.num_samples = self.msl.shape[0]

    @property
    def num_batches(self):
        return int(np.floor(self.num_images / self.batch_size))

    def __len__(self):
        return self.num_batches

    def min_max_normalize(self, arr, min, max):
        normalized = (arr - min) / (max - min)
        result = np.where(np.isnan(arr), np.nan, normalized)
        return result

    # must be called to restart the sequential
    def counter_reset(self):
        self.counter = 0

    def __getitem__(self, idx):

        # prepare the resulting array
        inputs = np.zeros((self.batch_size, 64, 64, 1))
        outputs = np.zeros((self.batch_size, 64, 64, 1))

        # random path
        if self.sequential == False:
            # compose the batch one element at the time
            for i in range(self.batch_size):
                # get a random number in range
                random = np.random.randint(0, self.num_samples - 1)

                # inputs
                inputs[i, :, :, 0] = self.min_max_normalize(
                    self.msl[random], 95680, 104256
                )

                # outputs
                outputs[i, :, :, 0] = self.min_max_normalize(
                    self.tp[random], 0.0, 0.02197266
                )

        # sequential path
        if self.sequential == True:
            # compose the batch one element at the time
            for i in range(self.batch_size):
                # inputs
                inputs[i, :, :, 0] = self.min_max_normalize(
                    self.msl[self.counter], 95680, 104256
                )

                # outputs
                outputs[i, :, :, 0] = self.min_max_normalize(
                    self.tp[self.counter], 0.0, 0.02197266
                )

                self.counter = self.counter + 1
        return inputs, outputs
